# (C) 2014-2015 Elke Schaper

"""
    :synopsis: The system paths module.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import os

# Where are the executables stored? E.g. the tandem repeat detection
# algorithms?
EXEC_DIR = "/usr/local/bin"

# What is the path to the package? E.g. path/to/Tral/tral
PACKAGE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))

# The default location of the local config files
CONFIG_DIR = os.path.join(os.path.expanduser('~'), ".tral")

# Where are the data-files, e.g. for pvalue calculation stored?
DATA_DIR = os.path.join(CONFIG_DIR, "data")


def config_file(name):

    ''' Returns the complete path to the config file `name`.

        Returns the complete path to the config file `name`. Searches first in
        `CONFIGDIR`. Otherwise, returns the distribution config file.
    '''

    if os.path.isfile(os.path.join(CONFIG_DIR, name)):
        return os.path.join(CONFIG_DIR, name)
    else:
        return os.path.join(PACKAGE_DIRECTORY, 'data', name)



import logging
import logging.config
import os
import pytest

from tral.paths import *

logging.config.fileConfig(config_file("logging.ini"))
log = logging.getLogger('root')

from tral.sequence import repeat_detection_run, sequence
from tral.hmm import hmm

TEST_FAA_FILE_MBE_2014 = "P51610.fasta"
TEST_HMM_FILES_MBE_2014 = ["Kelch_1.hmm", "Kelch_2.hmm"]
TEST_SCORE_MBE_2014 = "phylo_gap01_ignore_trailing_gaps_and_coherent_deletions"

# TEST RESULT MBE 2014
# all[all$Ensembl_Protein_ID == 'ENSP00000309555',]
# TR_ID lHMM lRepeat_Region lRepeat_Region_Mean lRepeat_Region_Max lMean  n        n_effective detection_ID
# 19571_1   47            306           257.00353                364  50.0  5  4.061224      PF01344
# 19571_0   49            313           274.15901                354  48.8  6  5.265306      PF07646
# 19571_3   28            426            69.73498                530  29.0 14 13.275862      xstream
# 19571_2   28            429            69.02473                533  28.0
# 14 13.500000      xstream

# RESULTS DE NOVO:
# T-REKS:
# ["QALAI-","QAV-L-","QAA-Q-","QAV-MG"]
# ["TALE--ALLCPSATVTQ-VCSNPPCETHETGTTN","TATT--SN-AGSAQR--V-CSNPPCETHETGTTH","TATTATSN-GGTGQP--EGGQQPPA-----GRPC"]
# ["TSTVGQQNGSVVRVC-SNPPCETHETGTTNTATTA","TSNMAGQHG-----C-SNPPCETHETGTTNTATTA","MSSVGANHQ-----RDARRACAAG-TPAVIRISVA"]
# TRUST
# ["T--AATA-TS-------PTP-NPvP-SVpANP-----P---KSPAP--A---A---------AA--PAVQPLTQVG----IT-L--L-PQAAP-AP-------PT--TT-TIQVLP","T--VPGS-SI-------SVP-TA-A-RT-QGV-----P--AVLKVT--GpqaT---------TG--TPLVTMRPAS----QAGK--ApVTVTS-LPagvrmvvPT--QSAQ-GTV-","T--LITA-PS-------GVE-AQ-P--V-HDL-----P--VSILAS--P---T---------TEqpTATVTIADSGqgdvQPGT--V-TLVCS-NP-------PC--ETHETGTTN","T--ATTTvVA-------NLG-GH-P-QP-TQV-----QfvCDRQEA--A------------------ASLVTSTVGq---QNGS--V-VRVCS-NP-------PC--ETHETGTTN","T--ATTA-TS-------NMA-GQ-H-GC-SNP-----P--CETHET--G---T---------TN--TATTAMSSVG----ANHQrdA-RRACA------------------AGTPA","VirISVA-TG-------ALE-AA-Q-G--SKS-----Q--CQTRQT--S---A---------TS--TTMTVMAT-G----AP---------CSaGPllg----PS--MAREPGGRS","P--AFVQ-LA-------PLS-SK-V-RL-SSPsikdlP--AGRH-S--H---A---------VS--TAAMTRSSVG-----AGE--P-RMA----P-------VC--ESLQGGSPS","T--TVTV-TAleallcpSAT-VT-Q-VC-SNP-----P--CETHET--G---T---------TN--TATT-----S----NAGS--A-QRVCS-NP-------PC--ETHETGTTH","T--ATTA-TS-------NGGtGQ-PeGG-QQPpagr-P--CETHQ-------T---------TS--TGTTMSVSV-------GA--L-L------P----------------DATS","S--HRTV-ES-------GLE-------V-AAA-----P--SVTPQA--G---T--------------------ALL----APFP--T-QRVCS-NP-------PC--ETHETGTTH","R--AVTT-VT-------QST-PV-P-GP-SVP-----P--PEELQVspG---PrqqlpprqlLQ--SASTAL--MG----ESAE--V-LSASQ-TPel-----PAavDLSSTGEPS"]
# ["SN-PATrmLKTA-AAQVGTSVSSATNTSTRPIITVHK----SGTVTVAQQAQVVTTV----VGG--VTKTITLV","KS-PIS--VPGG-SALISNLGKVMSVVQTKPVQTSAV----TGQASTGPVTQIIQTKGPLPAGT--ILKLVTSA","DGkPTT--IIT--TTQASGAGTKPTILGISSVSPST-----TKPGTTTIIKTIPMSAIITQAGAtgVTSS-PGI","KS-PIT--IITT-KVMTSGTGAPAKIITAVP---KIA----TGHGQQG-VTQVVLKGAPGQPGT--ILRTV---","---PM-----GG-VRLVT----PVTVSAVKPAVTTLVvkgtTGVTTLGTVTGTVSTSLAGAGGH--STS--ASL","AT-PIT--TLGTiATLSSQVINPTAITVSA-AQTTLT----AAGGLTTPTITMQPVSQPTQ-------------"]
# ["WKRVVG----W-SGPVPRPRHGHRAVA------IKELIVVFGggneG--------------I-------------------VDELhvYNTA----TNQWF","IPAVRGDIP------PGCAAYG--FVC------DGTRLLVFG----Gmveygk--------Y-------------------SNDL--YELQ----ASRWE","WKRLKAKTP-K-NGPPPCPRLGHSFSL------VGNKCYLFG----GlandsedpknniprY-------------------LNDL--YILElrpgSGVVA","W-----DIPiT-YGVLPPPRESHTAVVytekdnKKSKLVIYG----Gmsgc----------R-------------------LGDL--WTLD----IDTLT","WNK-----P-SlSGVAPLPRSLHSATT------IGNKMYVFG----G--------------WvplvmddvkvathekewkcTNTL--ACLN----LDTMA","WETILMDTL-E-DNI-PRARAGHCAVA------INTRLYIWS----G-----------------------------------------------------"]
# XSTREAM
# ["CSNPPCETHETGTTNTATTATSNMAG-QHG","CSNPPCETHETGTTNTATTAMS-SVGANHQ","CSNPPCETHETGTTNTATTAMSMAGNHG","CSNPPCETHETGTTNTATTAMS-MAG-NHG"]
# ["VCSNPPCETHETGTTNTA-T-TSNAGSAQR","VCSNPPCETHETGTTHTATTATSNGGTGQP","VCSNPPCETHETGTTHTATTSNAGSAQP","VCSNPPCETHETGTTHTA-T-TSNAGSAQP"]
# ["GTVT","GTVS","GTVS","GTVS"]
# ["AAAE","AAAQ","AAA","AAAE","AAAE"]
# ["ATA","ATA","ATA","ATA"]
# ["EGQ","EGQ","EGQ","EGQ"]
# ["A","A","A","A","A","A","A","A"]


TEST_FAA_FILE = "HIV-1_388796.faa"
TEST_HMM_FILE = "zf-CCHC.hmm"
TEST_DENOVO_PARAMETERS = {"detection": {"detectors": ["XSTREAM", "T-REKS"]}}
TEST_SCORE = "phylo_gap01"
TEST_RESULT_SEQ1 = [3,
                    [["LFNSTKLE", "LFNSST-N"],
                     ["GDII", "GDIR"], ["FLG", "FLG"]],
                    3,
                    [["FNCGG-EF", "FYCNTSNL", "FNSTKLEL", "FNSST-NL"],
                        ["GDII", "GDIR"], ["FLG", "FLG"]]]

notfixed = pytest.mark.notfixed


@pytest.fixture
def path():
    """Return the path to the test data files.
    """
    return os.path.join(os.path.abspath('.'), 'test')


@pytest.mark.slow
def test_MBE_2014_workflow():
    # The Schaper et al. (MBE, 2014) workflow is tested on a single sequence.

    test_lSeq = sequence.Sequence.create(
        os.path.join(
            path(),
            TEST_FAA_FILE_MBE_2014),
        input_format='fasta')
    test_seq = test_lSeq[0]

    # Information on sequence domains (here: Pfam) in this sequence are added.
    test_pfam_hmm = [
        hmm.HMM.create(
            input_format='hmmer',
            file=os.path.join(
                path(),
                i)) for i in TEST_HMM_FILES_MBE_2014]

    # The sequence is searched for tandem repetitions of the Pfam domain in
    # the sequence
    test_pfam_list = test_seq.detect(lHMM=test_pfam_hmm)
    assert len(test_pfam_list.repeats) == 2

    # Pfam TRs with n_effective < 3.5 are discarded.
    test_pfam_list = test_pfam_list.filter("attribute", "n_effective", "min", 3.5)
    assert len(test_pfam_list.repeats) == 2

    # de novo detection methods (Trust, T-reks, Xstream, HHrepID) are used to
    # search the
    test_denovo_list = test_seq.detect(denovo=True, **TEST_DENOVO_PARAMETERS)
    # When Trust is part of the detectors, the number of found repeats may
    # differ between runs...
    assert len(test_denovo_list.repeats) == 10

    # De novo TRs with dTR_units (divergence) > 0.8; n_effective < 2.5; l < 10
    # or pvalue "phylo_gap01_ignore_trailing_gaps_and_coherent_deletions" >0.01
    # are discarded.
    test_denovo_list = test_denovo_list.filter(
        "pvalue",
        TEST_SCORE_MBE_2014,
        0.01)
    assert len(test_denovo_list.repeats) == 10
    test_denovo_list = test_denovo_list.filter(
        "divergence",
        TEST_SCORE_MBE_2014,
        0.8)
    assert len(test_denovo_list.repeats) == 10
    test_denovo_list = test_denovo_list.filter("attribute", "n_effective", "min", 2.5)
    assert len(test_denovo_list.repeats) == 5
    test_denovo_list = test_denovo_list.filter("attribute", "l_effective", "min", 10)
    assert len(test_denovo_list.repeats) == 2

    # De novo TRs were remastered with HMM
    test_denovo_hmm = [
        hmm.HMM.create(
            input_format='repeat',
            repeat=iTR) for iTR in test_denovo_list.repeats]
    test_denovo_list_remastered = test_seq.detect(lHMM=test_denovo_hmm)
    assert len(test_denovo_list_remastered.repeats) == 2

    # pvalue "phylo_gap01_ignore_trailing_gaps_and_coherent_deletions" > 0.1
    # are discarded.
    test_denovo_list_remastered = test_denovo_list_remastered.filter(
        "pvalue",
        TEST_SCORE_MBE_2014,
        0.1)

    # De novo TRs were filtered (n_effective < 3.5 are discarded.)
    test_denovo_list_remastered = test_denovo_list_remastered.filter(
        "attribute",
        "n_effective",
        "min",
        3.5)
    assert len(test_denovo_list_remastered.repeats) == 2

    # De novo TRs overlapping with a Pfam TR were filtered
    test_denovo_list_remastered = test_denovo_list_remastered.filter(
        "none_overlapping_fixed_repeats",
        test_pfam_list,
        "shared_char")
    assert len(test_denovo_list_remastered.repeats) == 2

    # Remaining De novo TRs were clustered for overlap (common ancestry). Only best =
    # lowest p-Value and lowest divergence were retained.
    test_denovo_list_remastered = test_denovo_list_remastered.filter(
        "none_overlapping", ("common_ancestry", None), [
            ("pvalue", TEST_SCORE_MBE_2014), ("divergence", TEST_SCORE_MBE_2014)])
    assert len(test_denovo_list_remastered.repeats) == 1

    # Merge remaining set of de novo and Pfam TRs.
    test_entire_set = test_pfam_list + test_denovo_list_remastered
    assert len(test_entire_set.repeats) == 3

    # Write result set of Pfam TRs
    #test_entire_set.write(format = "tsv, ...")



# (C) 2012-2015 Elke Schaper
# coding: utf-8

import math
import numpy as np
import scipy as sp
import scipy.stats
import scipy.special
import scipy.linalg
from scipy.optimize import fsolve, minimize_scalar
import logging
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.ERROR)

from tral.repeat import repeat
from tral.repeat.repeat_score import load_model
from tral.hmm import hmm_io

################################### HMM class ############################


class HMM:

    """ A cyclic HMM applicable to describe sequence Tandem Repeats """

    def __init__(
            self,
            tandem_repeat=None,
            prior_divergence=None,
            prior_indel_insertion=None,
            parameters=None):

        if tandem_repeat:
            if isinstance(tandem_repeat, str):
                self.HMM_from_file(
                    tandem_repeat,
                    accession=parameters,
                    prior_indel_insertion=prior_indel_insertion)
            else:
                self.HMM_from_TR(
                    tandem_repeat,
                    prior_divergence=prior_divergence,
                    prior_indel_insertion=prior_indel_insertion,
                    emission_file=parameters)
        else:
            self.HMM_example()

    def HMM_example(self):
        #states = ["N", "B", "M1", "M2", "M3", "E", "C"]
        self.states = ["N", "M1", "M2", "C"]

        # Initialisation
        self.p_t = {iS: {iS2: 0 for iS2 in self.states}
                    for iS in self.states[:-1]}
        self.p_0 = {iS: 1 / 3 for iS in self.states}

        # Transition probabilities
        # Feed Values to p_t
        self.p_t["N"] = {"N": 0.5, "M1": 0.5}
        self.p_t["M1"] = {"M2": 0.5, "C": 0.5}
        self.p_t["M2"] = {"M1": 0.5, "C": 0.5}
        self.p_t["C"] = {"C": 1}

        # emissions
        self.emissions = ["A", "C", "G", "T"]

        # emission probabilities
        self.p_e = {iS: {iE: 0.25 for iE in self.emissions}
                    for iS in self.states}
        self.p_e['M1'] = {"A": 0.9, "C": 0.025, "G": 0.025, "T": 0.025}
        self.p_e['M2'] = {"A": 0.025, "C": 0.9, "G": 0.025, "T": 0.025}

    def HMM_from_TR_One_step(self, tandem_repeat, t):
        """ Build a HMM from a TR alignment given the ML divergence t.
            One step approach: Calc the posterior of the ancestral sequence and use it
            as the likelihood of the homologous repeat unit sequence. """
        # not implemented yet. For implementation, Use the first half of
        # self.HMM_from_TR()

    def HMM_from_file(self, hmm_file, accession, prior_indel_insertion):
        """ Load a HMM from hmm_file
            Store probabilities as log10arithms.

            Parameters:
            <hmm_file> = 'path//to/file.hmm'
            <accession> = 'PF00560'
            <prior_indel_insertion> = {'mu': 0.5, 'sigma_squared': 0.81}


        """

        bInsertion_state_emission_p_from_file = False
        bTerminal_state_emission_p_from_file = False

        hmmer_probabilities = hmm_io.read_HMMER(hmm_file, id=accession)
        dTranslate_States = {
            i: str(
                int(i) -
                1) for i in hmmer_probabilities.keys() if i not in [
                'COMPO',
                'letters']}
        hmmer_probabilities = {(i if i in ['COMPO', 'letters'] else dTranslate_States[
                                i]): j for i, j in hmmer_probabilities.items()}
        # WARNING: The order of keys in hmmer_probabilities might not be what
        # you think. You need to adapt your code at several instances (wherever
        # you assume the states to be ordered, e.g. zip(repeat_states,
        # hmmer_probabilities.keys))

        if not hmmer_probabilities:
            print("<hmmer_probabilities> were not found :(")
            return None

        self.hmmer = hmmer_probabilities

        # Warning! The following line is only appropriate, if there are only
        # "letters", "COMPO", and the match state keys in <hmmer_probabilities>
        l_effective = len(hmmer_probabilities.keys()) - 2
        # Assume: sequence_type = 'AA'
        Q, null_model_emission_p, alphabet = load_model('lg')

        # Initialise all HMM states to default value
        repeat_states, insert_states = self.initialise_HMM_structure(l_effective)

        # Store indel probabilities
        self.set_indel_probabilities(prior_indel_insertion, l_effective, default=True)

        # Store null model insertion probabilities for terminal and insertion
        # states
        if bTerminal_state_emission_p_from_file:
            null_model_emission_p = hmmer3_null_model_emission_probabilites(
                hmmer_probabilities=hmmer_probabilities)

        self.hmmer_probabilities = hmmer_probabilities

        # Store match state and insertion state emission probabilities
        match_state_emission_p = hmmer3_emission_probabilities(
            hmmer_probabilities=hmmer_probabilities,
            letters=list(
                alphabet.keys()),
            lMatch=repeat_states)
        if bInsertion_state_emission_p_from_file:
            insertion_state_emission_p = hmmer3_insertion_probabilities(
                hmmer_probabilities=hmmer_probabilities,
                letters=list(
                    alphabet.keys()),
                lMatch=insertion_states)
        else:
            insertion_state_emission_p = None
        self.set_emission_probabilities(
            alphabet,
            null_model_emission_p,
            match_state_emission_p,
            repeat_states,
            insertion_state_emission_p,
            insert_states)

    def HMM_from_TR(
            self,
            tandem_repeat,
            prior_divergence=None,
            prior_indel_insertion=None,
            emission_file=None):
        """ Build a HMM from a TR <tandem_repeat>.
            Two step approach: First, calc the posterior of the ancestral sequence
            Then, calculate the likelihood of the homologous repeat unit sequence.

            Parameters:
            <prior_divergence> is {'type': 'alpha'/'fixed_value', 'value': 4}
            <prior_indel_insertion> is {'mu': 0.5, 'sigma_squared': 0.81}

            Store probabilities as log10arithms."""

        # This will fail at current, as the divergence attribute is a dict.
        if prior_divergence is None:
            divergence = divergence_from_FP_simulations(tandem_repeat.l_effective)
        elif prior_divergence['type'] == 'alpha':
            divergence = divergence_from_FP_simulations(
                tandem_repeat.l_effective,
                alpha=prior_divergence['value'])
            # print(divergence)
            # print(prior_divergence['value'])
        else:
            divergence = prior_divergence['value']

        if prior_indel_insertion is not None and prior_indel_insertion[
                'type'] == 'adaptive':
            # Adapt the indel probabilities towards the divergences
            prior_indel_insertion['mu'] = divergence / \
                prior_indel_insertion['factor']
            prior_indel_insertion[
                'sigma_squared'] = prior_indel_insertion['mu'] / 5

        sequence_type = tandem_repeat.sequence_type
        msaTD = tandem_repeat.msaTD
        l_effective = tandem_repeat.l_effective
        n = tandem_repeat.n  # Test: Do you really need this variable?

        repeat_states, insert_states = self.initialise_HMM_structure(l_effective)

        # Transitions incorporating deletion or insertion states

        # We assume that deletions are not longer than the repeat unit, which would hence not be visible anymore.
        # We also assume that you cannot jump from "I" directly to "C", as is
        # would have been more likely to directly assume "C" from the foregoing
        # state.

        # # Apply the modulus in order to determine the true index of any state,
        # as you might have crossed the current tandem repeat unit's border:
        # index(i,l_effective) = i%l_effective

        insertion_lengths, deletion_lengths = tandem_repeat.gap_structure_HMM()
        self.set_indel_probabilities(
            prior_indel_insertion,
            l_effective,
            n=n,
            default=False,
            insertion_lengths=insertion_lengths,
            deletion_lengths=deletion_lengths)

        # Load model of sequence evolution parameters
        if sequence_type == 'AA':
            Q, eqFreq, alphabet = load_model('lg')
        else:
            Q, eqFreq, alphabet = load_model('tn93')

        # Calculate ML-emission probabilities for all match states
        # YOU might want to transfer Q,eqFreq,alphabet to this function???
        if not emission_file:
            likelihood_offspring = calculate_log10_offspring_likelihood(
                tandem_repeat,
                divergence)
        else:
            hmmer_probabilities = hmm_io.read_HMMER(emission_file)[0]
            likelihood_offspring = hmmer3_emission_probabilities(
                hmmer_probabilities=hmmer_probabilities,
                letters=list(
                    alphabet.keys()),
                lMatch=repeat_states)

        # Anpassen!
        self.set_emission_probabilities(
            alphabet=alphabet,
            null_model_emission_p=eqFreq,
            match_state_emission_p=likelihood_offspring,
            repeat_states=repeat_states)

    def initialise_HMM_structure(self, l_effective):
        '''
        Initialise all states
        Set transition probabilities to None/0
        Set initial probability for all states equal.
        '''

        # Build a HMM from these likelihoods.
        repeat_states = ["M{0}".format(str(i)) for i in range(l_effective)]
        insert_states = ["I{0}".format(str(i)) for i in range(l_effective)]
        self.states = ["N"] + repeat_states + insert_states + ["C"]
        logger.debug("HMM states: {0}".format(self.states))

        # Initialisation
        # The transition probability is initially set to None for all states.
        self.p_t = {iS: {iS2: None for iS2 in self.states}
                    for iS in self.states}
        # The initial probability is equal in all states. (np.log10(1)=0)
        self.p_0 = {iS: 0 for iS in self.states}

        # Transition probabilities

        # First, mark which transitions are not penalised, e.g. bear a 'probability of np.log10(1)=0':
        # - Transitions from "N" to any "M_k"
        # - Transitions from any "M_k" to "C"
        # - Transitions from "M_k" to "M_k+1" (for k<n) and "M_k" to "M_0" (for k==n)
        # - Transitions from "N" to "N"
        # - Transitions from "C" to "C"
        for i in range(l_effective):
            self.p_t["N"]["M{0}".format(index(i, l_effective))] = 0
            self.p_t["M{0}".format(index(i, l_effective))]["C"] = 0
            self.p_t[
                "M{0}".format(
                    index(
                        i,
                        l_effective))][
                "M{0}".format(
                    index(
                        i +
                        1,
                        l_effective))] = 0
        self.p_t["N"]["N"] = 0
        self.p_t["C"]["C"] = 0

        return repeat_states, insert_states

    def set_emission_probabilities(
            self,
            alphabet,
            null_model_emission_p,
            match_state_emission_p,
            repeat_states,
            insertion_emmission_p=None,
            insertion_states=None):

        # emissions
        self.emissions = alphabet.keys()
        logger.debug("Emissions: {0}".format(self.emissions))

        # emission probabilities

        # Set all emission probabilities to null model emission probabilities
        self.p_e = {
            iS: {
                letter: np.log10(
                    null_model_emission_p[number]) for letter,
                number in alphabet.items()} for iS in self.states}

        # Set match state probabilities to ML-emission probabilities
        for iRS, iEmission_p in zip(repeat_states, match_state_emission_p):
            self.p_e[iRS] = iEmission_p

        if insertion_emmission_p and insertion_states:
            print("Setting insertion state probabilities")
            for iIS, iEmission_p in zip(
                    insertion_states, insertion_emmission_p):
                self.p_e[iIS] = iEmission_p

        #logger.debug("Emission probabilities: {0}".format(self.p_e['M0']))

    def set_indel_probabilities(
            self,
            prior_indel_insertion,
            l_effective,
            n=2,
            default=True,
            insertion_lengths=None,
            deletion_lengths=None):

        if default:
            p_deletion_formation = calculate_log10_indel_probability(
                0,
                n,
                prior=prior_indel_insertion)
            p_insertion_formation = calculate_log10_indel_probability(
                0,
                n,
                prior=prior_indel_insertion)
            p_deletion_lengths = calculate_log10_probability_indel_lengths(
                [],
                l_effective -
                1,
                type='zipf')
            p_insertion_continued, p_insertion_stopped = calculate_log10_probability_indel_lengths(
                [], None, type='exponential')

        # First, calculate the transition probability of all states that only
        # depends on the gap structure of one site:
        for i in range(l_effective):

            if not default:
                p_deletion_formation = calculate_log10_indel_probability(
                    len(deletion_lengths[i]), n, prior=prior_indel_insertion)
                p_insertion_formation = calculate_log10_indel_probability(
                    len(insertion_lengths[i]), n, prior=prior_indel_insertion)
                p_deletion_lengths = calculate_log10_probability_indel_lengths(
                    deletion_lengths[i],
                    l_effective -
                    1,
                    type='zipf')
                p_insertion_continued, p_insertion_stopped = calculate_log10_probability_indel_lengths(
                    insertion_lengths[i], None, type='exponential')

            # In the current setup, moving from "N" to either an insertion, or a deletion state
            # is less likely than staying in "N". Thus, these probabilities are left at zero.
            #self.p_t["N"]["I{0}".format(str(i))] = p_insertion

            # Mi-1 -> Mi+l_effective (deletion(l_effective))
            for iC, p_deletion_length in enumerate(p_deletion_lengths):
                self.p_t[
                    "M{0}".format(
                        index(
                            i - 1,
                            l_effective))][
                    "M{0}".format(
                        index(
                            i + iC + 1,
                            l_effective))] = p_deletion_formation + p_deletion_length
                # Ii -> Mi + l_effective (insertionL) + deletion(l_effective)
                self.p_t[
                    "I{0}".format(
                        index(
                            i,
                            l_effective))][
                    "M{0}".format(
                        index(
                            i + iC + 1,
                            l_effective))] = p_insertion_stopped + p_deletion_formation + p_deletion_length

            # Mi-1 -> Ii (insertion)
            self.p_t[
                "M{0}".format(
                    index(
                        i - 1,
                        l_effective))][
                "I{0}".format(
                    index(
                        i,
                        l_effective))] = p_insertion_formation

            # Ii -> Ii (1-insertionL)
            self.p_t[
                "I{0}".format(
                    index(
                        i,
                        l_effective))][
                "I{0}".format(
                    index(
                        i,
                        l_effective))] = p_insertion_continued

            # Ii -> Mi (insertionL)
            self.p_t[
                "I{0}".format(
                    index(
                        i,
                        l_effective))][
                "M{0}".format(
                    index(
                        i,
                        l_effective))] = p_insertion_stopped

        # Second, calculate the transition probability of all states that depends on the gap structure of several sites.
        # These calcs use the information already gathered in self.p_t.
        # These are deletions, that are directly followed by insertions
        # (Case 1: deletion from match state. Case 2: deletion from insertion state.)
        for i in range(l_effective):
            for iC in range(1, l_effective):
                # Mi -> Ii+iC (deletion(iC-1) + insertion(i+iC))
                self.p_t[
                    "M{0}".format(
                        index(
                            i,
                            l_effective))][
                    "I{0}".format(
                        index(
                            i + iC,
                            l_effective))] = self.p_t[
                    "M{0}".format(
                        index(
                            i,
                            l_effective))][
                    "M{0}".format(
                        index(
                            i + iC,
                            l_effective))] + self.p_t[
                    "M{0}".format(
                        index(
                            i + iC - 1,
                            l_effective))][
                    "I{0}".format(
                        index(
                            i + iC,
                            l_effective))]

                # Ii -> Ii+l_effective insertion_stop(i) + deletion(i;l_effective) +
                # insertion_formation(i+l_effective+1)
                self.p_t["I{0}".format(index(i,
                                             l_effective))]["I{0}".format(index(i + iC,
                                                                       l_effective))] = self.p_t["I{0}".format(index(i,
                                                                                                            l_effective))]["M{0}".format(index(i,
                                                                                                                                      l_effective))] + self.p_t["M{0}".format(index(i - 1,
                                                                                                                                                                           l_effective))]["M{0}".format(index(i + iC,
                                                                                                                                                                                                     l_effective))] + self.p_t["M{0}".format(index(i + iC - 1,
                                                                                                                                                                                                                                          l_effective))]["I{0}".format(index(i + iC,
                                                                                                                                                                                                                                                                    l_effective))]

        logger.debug("Transition probabilities: {0}".format(self.p_t))


################################### Local TR class #######################
class TR:

    """ Fake interim TR class """

    def __init__(self):

        self.divergence = 0.6
        self.msaTD = ["AAA", "CCC", "GTG"]
        self.sequence_type = 'DNA'
        self.n = 3
        self.l_effective = 3

################################### MAP parameter estimation 3############


def divergence_from_FP_simulations(l, alpha=0.1):
    ''' Which HMM divergence sets the average FP-rate
        (i.e. the number of falsely positively assigend amino acid on either flanking side to a perfect TR)
        to in total alpha*l'''

    # alpha: In average, the number of falsely assigned AAs on both sides of the TR with should not extend l * alpha
    # Derivation:
    # FP-rate (average number of AAs predicted falsely as part of a perfect
    # TR) on one side of the TR as a function of the divergence.
    f = [
        0.50150,
        0.58600,
        0.70900,
        0.75125,
        0.89275,
        1.00900,
        1.25100,
        1.24150,
        1.56275,
        1.69600,
        1.88475,
        2.20025,
        2.33375,
        2.53500,
        2.99300,
        3.25900,
        3.77400,
        4.03300,
        4.26675,
        4.71800,
        5.30925,
        5.83325,
        6.31575,
        6.84000,
        7.93425,
        9.01900,
        9.72000,
        10.80925,
        12.26650,
        13.50750,
        14.29825,
        15.74225,
        18.66050,
        20.02575,
        21.54400,
        23.97300,
        26.93625,
        30.39400,
        32.78550,
        35.80375,
        38.98025,
        45.79925,
        48.48200,
        51.29100,
        56.74300,
        60.45925,
        64.44675,
        69.41900,
        80.02450,
        82.37550,
        88.47350,
        95.13425,
        98.00750,
        105.54975,
        108.85625,
        117.76750,
        125.69250,
        125.73725,
        137.40675,
        143.08225,
        148.28725,
        154.51425,
        161.23075,
        160.33875,
        172.00475,
        175.41525,
        180.44425,
        179.47350,
        190.14225,
        190.93625,
        194.03725,
        194.87475,
        203.35900,
        209.94350,
        210.63750,
        210.63950,
        216.69850,
        223.15025,
        219.22775,
        227.37700,
        227.09600,
        238.11900,
        235.79275,
        238.05250,
        242.59875,
        244.40425]
    # Sequence divergence associated with f.
    d = [(i / 10) + 0.5 for i in range(86)]
    # List <da> contains the linearly weighted averaged divergences of the two
    # values of the FP-rate that enclose l*alpha:
    da = []
    before = (f[0], d[0])
    for iF, iD in zip(f[1:], d[1:]):
        f_ok = (alpha * (len(da) + 1)) / 2
        while iF > f_ok:
            da.append(
                ((iF - f_ok) * before[1] + (f_ok - before[0]) * iD) / (iF - before[0]))
            f_ok = (alpha * (len(da) + 1)) / 2
        before = (iF, iD)

    if l > len(da):
        return 6
    else:
        return da[l - 1]


def index(i, iMax):
    return str(i % iMax)


def calculate_log10_offspring_likelihood(tandem_repeat, divergence=None):

    sequence_type = tandem_repeat.sequence_type
    msaTD = tandem_repeat.msaTD
    l_effective = tandem_repeat.l_effective
    n = tandem_repeat.n

    # Calculate posterior probabilities of ancestral states from the TR
    # alignment

    # Load model of sequence evolution parameters
    if sequence_type == 'AA':
        Q, eqFreq, alphabet = load_model('lg')
    else:
        Q, eqFreq, alphabet = load_model('tn93')
    P = scipy.linalg.expm(Q * divergence)
    alphabet_reverse = {i: j for j, i in alphabet.items()}

    # Initialise the count matrix (if not already existent)
    try:
        tandem_repeat.msaTDN
    except:
        msaTDN_temp = [[alphabet[symbol] for symbol in column if symbol not in [
            '-', 'X', '*', 'U']] for column in tandem_repeat.msaTD]
        tandem_repeat.msaTDN = []
        for column in msaTDN_temp:
            my_diag = {i: 0 for i in range(len(alphabet.keys()))}
            for iN in column:
                my_diag[iN] = my_diag[iN] + 1
            tandem_repeat.msaTDN.append(
                np.array([j for j in my_diag.values()]))

    # Initialise results matrix
    posterior_ancestor = []

    # Calculate the posterior probabilities per column
    for column in tandem_repeat.msaTDN:
        posterior = {
            i: iJ *
            np.product(
                np.power(
                    P[i],
                    column)) for i,
            iJ in enumerate(eqFreq)}
        normalising_factor = sum(posterior.values())
        for key, value in posterior.items():
            posterior[key] = value / normalising_factor
        posterior_ancestor.append(posterior)

    # Calculate the likelihood of evolved TR units with the given divergence
    # from these ancestral states
    likelihood_offspring = [
        {
            iA: np.log10(
                sum(
                    posterior[iAncestor] *
                    P[iAncestor][i] for iAncestor in alphabet.values())) for iA,
            i in alphabet.items()} for posterior in posterior_ancestor]
    return likelihood_offspring


def loglikelihood_substitution(t, Q, eqFreq, alphabet, tandem_repeat):
    ''' calculate the likelihood of a repeat assuming a star tree and a sequence model
        defined by Q, t, eqFreq and alphabet. '''
    P = scipy.linalg.expm(Q * t)

    # msaTDN is a list of tandem_repeat.n numpy arrays.
    # Each numpy array represents the elements of the alphabet. The entries
    # are the count of an alphabet element in the respective tandem_repeat
    # column
    try:
        tandem_repeat.msaTDN
    except:
        msaTDN_temp = [[alphabet[symbol] for symbol in column if symbol not in [
            '-', 'X', '*', 'U']] for column in tandem_repeat.msaTD]
        tandem_repeat.msaTDN = []
        for column in msaTDN_temp:
            my_diag = {i: 0 for i in range(len(alphabet.keys()))}
            for iN in column:
                my_diag[iN] = my_diag[iN] + 1
            tandem_repeat.msaTDN.append(
                np.array([j for j in my_diag.values()]))

    # First sum: Over all sites
    # Second sum: Over all possible ancestral characters
    # Third product: Over all repeat units
    likelihood = sum(
        np.log(
            np.sum(
                iJ *
                np.product(
                    np.power(
                        P[i],
                        column)) for i,
                iJ in enumerate(eqFreq))) for column in tandem_repeat.msaTDN)
    return likelihood


def calculate_log10_probability_indel_lengths(
        indel_lengths,
        indel_length_max,
        type='zipf',
        prior=None):
    ''' calculate the probability of indels of different length until a maximum length <indel_length_max>,
    assuming that either

    type == 'zipf'
    indel lengths are distributed following a Zipfian distribution with parameter <indel_zipf>.
    (Compare Fletcher,W. and Yang,Z. (2009) INDELible: a flexible simulator of biological sequence evolution. Mol Biol Evol, 26, 1879–1888.
    In their publication, <indel_zipf> is denoted as <a>, the gap lengths as <u>.)

    or
    type == 'exponential'
    indels lengths are distributed following a (1-alpha)*(alpha^(l-1)) distribution.
    [Is this model already used? Probably yes. How can i cite it?]

    A good choice for <indel_length_max> when building HMMs for tandem repeats is <l_effective-1>, as a deletion, that
    includes a complete repeat unit, will not be visible in the tandem repeat unit alignment, and could therefore
    be ignored.

    Return the probabilities as a list.
    '''

    if 'zipf' == type:
        indel_zipf = calculate_MAP_Indel_length_Zipfian_factor(
            indel_lengths,
            prior)
        zeta_factor = np.log10(1 / scipy.special.zeta(indel_zipf, 1))
        return [
            np.log10(
                ((iL + 1) ** -indel_zipf)) + zeta_factor for iL in range(indel_length_max)]

    elif 'exponential' == type:
        alpha = calculate_MAP_Indel_length_exponential_factor(
            indel_lengths,
            prior)
        return np.log10(alpha), np.log10(1 - alpha)

    else:
        logger.error(
            "I am not aware of indel length model type: {0}.".format(type))


def calculate_MAP_Indel_length_exponential_factor(indel_lengths, prior=None):
    ''' calculate the MAP Exponential decay constant <alpha> that determines the distribution of
    indel lengths. Assume a Gaussian prior. Input is a list of <indel_lengths> for a particular
    column.

    The MAP is calculated numerically, as no nice analytical solution has been found so far [power function of grade 3 needs to be solved:
    0 == prior['sigma_squared']*(sum(indel_lengths) - len(indel_lengths)) - alpha*(prior['sigma_squared']*sum(indel_lengths) + prior['mu']) + (alpha**2)*(1+prior['mu']) - alpha**3
    '''

    if prior is None:
        prior = {'mu': 0.8, 'sigma_squared': 0.1}

    if len(indel_lengths) == 0:
        return prior['mu']

    else:
        # As we are only interested in the maximum, we can leave out factors (e.g. * 1/(math.sqrt(2*math.pi * prior['sigma_squared'])) for the Gaussian)
        # There is an analytic solution. Compare "Gap it!" or "gaps.nb"
        posterior = lambda alpha: - np.prod([(1 - alpha) * (alpha ** (lIndel - 1))
                                             for lIndel in indel_lengths]) * math.exp(-0.5 * ((alpha - prior['mu']) ** 2) / prior['sigma_squared'])
        res = minimize_scalar(posterior, bounds=(0, 1), method='bounded')
        return res.x


def calculate_MAP_Indel_length_Zipfian_factor(indel_lengths, prior=None):
    ''' calculate the MAP Zipfian constant <indel_zipf> that determines the distribution of
    indel lengths. Assume a Gaussian prior. Input is a list of <indel_lengths> for a particular
    column.
    The probability distribution of indel lengths is assumed to follow the Zipfian distribution
    (Compare Fletcher,W. and Yang,Z. (2009) INDELible: a flexible simulator of biological sequence evolution. Mol Biol Evol, 26, 1879–1888.
    In their publication, <indel_zipf> is denoted as <a>, the gap lengths as <u>.)

    The MAP is calculated numerically, as there did not seem to be a nice analytical solution.
    '''

    if prior is None:
        prior = {'mu': 1.821, 'sigma_squared': 1}

    if len(indel_lengths) == 0:
        return prior['mu']

    else:
        # The posterior is the likelihood * prior * scaling.
        # As we are only interested in the maximum, we leave out all factors in
        # the following equation for the posterior:
        posterior = lambda indel_zipf: - (scipy.special.zeta(indel_zipf, 1) ** - len(indel_lengths)) * np.prod(
            [lIndel ** -indel_zipf for lIndel in indel_lengths]) * math.exp(-0.5 * ((indel_zipf - prior['mu']) ** 2) / prior['sigma_squared'])
        res = minimize_scalar(posterior, method='brent')
        return res.x


def calculate_log10_indel_probability(nIndels, n, prior=None):
    ''' calculate the probabilty of an indel per site and per repeat unit
    from the MAP estimate of the indel rate, given
    - the <divergence> of the repeat units
    - a <prior> on the indel rates (normal distributed)
    - the number of indels <nIndels> in this column
    - the total length <n> of this column
    '''

    indel_rate_MAP = calculate_MAP_indel_rate(nIndels, n, prior)
    return np.log10(1 - math.exp(-indel_rate_MAP))


def calculate_MAP_indel_rate(nIndels, n, prior=None):
    ''' calculate the MAP indel_rate per repeat unit and site given
    - an assumed prior distribution of the indel rate described by the dict <prior>
    - the likelihood of the the observed number of indels <nIndels>.

    posterior ~ likelihood * prior
    Derive the posterior with respect to the indel_rate
    Set the result equal to 0
    Solve with respect to the MAP of the indel_rate.

    For nIndels in {0,1} it is possible to derive the MAP indel_rate analytically.
    For all other cases, the maximum of the posterior distribution needs to be found in good approximation algorithmically.

    ln(posterior) ~ ln(likelihood) + ln(prior)
    '''

    if prior is None:
        prior = {'mu': 0.01, 'sigma_squared': 0.001}

    if nIndels == 0:
        # This max is chosen absolutely randomly. do you have any better idea?
        # (Before: constant 0.001)
        return max(prior['mu'] / 100, prior['mu'] - prior['sigma_squared'] * n)
    else:
        # Calculate the roots numerically
        res = fsolve(
            derivative_log_posterior,
            prior['mu'],
            args=(
                prior,
                n,
                nIndels))
        return res[0]


def derivative_log_posterior(indel_rate, prior, n, nIndels):
    ''' assuming that on each site on each repeat unit, an in/del event has occured, or not.
    Thus, the likelihood of a certain number of in/del events follows the binomial distribution.
    The prior is assumed to be a gaussian described by prior.

    Return the derivative with respect to <indel_rate> of the natural logarithm of the posterior
    probability of <nIndels>, calculated as the product of likelihood and prior. The
    normalisation factor is ignored for the moment.'''
    return indel_rate + (prior['mu'] - indel_rate - n * prior['sigma_squared']) * \
        math.exp(-indel_rate) - prior['mu'] + (n - nIndels) * prior['sigma_squared']

############################### External parameters ######################


def hmmer3_emission_probabilities(hmmer_probabilities, letters, lMatch):
    '''
    Get emission probabilities from hmmer3 hmm file.
    In hmm file, emission probabilities are -ln(p).
    Return log10(p), i.d. convert between the two. Conversion: p_Local = - p_HMM * log10(e)

    Parameters (e.g.):
    letters = ['A', 'C', 'E', 'D', 'G', 'F', 'I', 'H', 'K', 'M', 'L', 'N', 'Q', 'P', 'S', 'R', 'T', 'W', 'V', 'Y']
    lMatch = ['M'+str(i) for i in range(24)]

    Return format (pseudo code):
    [{iA: np.log10(p(iA,iM)) for iA in alphabet.keys()} for iM in lMatch]

    '''

    # Test: Is the number of match states in both models equal?
    if not len(hmmer_probabilities.keys()) - 2 == len(lMatch):
        print('Match states HMMER: {0} Match states local: {1}'.format(
            len(hmmer_probabilities.keys()) - 2, len(lMatch)))
        raise ValueError(
            'The number of match states in HMMer model and local model does not match')

    # Test: Are all <letters> represented in the HMMER HMM?
    if any(iL not in hmmer_probabilities['letters'] for iL in letters):
        missing = [
            iL for iL in letters if iL not in hmmer_probabilities['letters']]
        print('Missing representation in Hmmer File: {0}'.format(missing))
        raise ValueError(
            'Some letters in the local HMM are not represented in the HMMER HMM.')

    return [
        {
            iL: -
            float(iP) *
            np.log10(
                np.exp(1)) for iL,
            iP in zip(
                hmmer_probabilities['letters'],
                hmmer_probabilities[
                    iMatch[
                        1:]]['emissions']) if iL in letters} for iMatch in lMatch]


##################################### Tests ##############################


def test():
    ''' To be implemented... '''
    #tandem_repeat = ...
    divergence = 0
    calculate_log10_offspring_likelihood(tandem_repeat, divergence)

##################################### Main ###############################


def main():
    my_TR = repeat_info.Repeat(
        begin=0,
        msa=[
            'A-G',
            'ACG',
            'ACG'],
        sequence_type='DNA')
    #my_TR = TR()
    my_HMM = HMM(my_TR, divergence=0.1)

    #print(calculate_indel_rate(1, 6, 1.5))

    # print(calculate_probability_indel_lengths(0.366869,6,'exponential'))

if __name__ == "__main__":
    main()



import numpy as np
import os
import pytest

from tral.repeat import repeat
from tral.repeat import repeat_io


TEST_MSA_O = ['OCC', 'OOO']
TEST_MSA_K = ['KCC', 'KKK']
TEST_SCORE = "phylo_gap01"

# defaultdict(<class 'int'>, {'pSim': 0.66666666669999997, 'parsimony': 0.66666666669999997, 'entropy': 0.66666666666666663, 'phylo': 0.11368675605567802})
# pvalue 'phylo': 0.3821

notfixed = pytest.mark.notfixed

@pytest.fixture
def path():
    """Return the path to the test data files.
    """
    return os.path.join(os.path.abspath('.'), 'repeat', 'test')


@pytest.mark.no_external_software_required
def test_standardize_amino_acids():

    assert repeat.standardize("ABDEF-G") == "ADDEF-G"


@pytest.mark.no_external_software_required
def test_repeat_ambiguous():

    myTR_O = repeat.Repeat(msa = TEST_MSA_O)
    myTR_K = repeat.Repeat(msa = TEST_MSA_K)

    assert myTR_O.msaTD_standard_aa == myTR_K.msaTD
    assert myTR_O.msaTD_standard_aa == myTR_K.msaTD_standard_aa

    assert myTR_O.score(TEST_SCORE) == myTR_K.score(TEST_SCORE)
    assert myTR_O.divergence(TEST_SCORE) == myTR_K.divergence(TEST_SCORE)
    assert myTR_O.pvalue(TEST_SCORE) == myTR_K.pvalue(TEST_SCORE)

    assert myTR_O.divergence(TEST_SCORE) == 2.095947265625
    assert myTR_K.pvalue(TEST_SCORE) == 0.3507


@pytest.mark.no_external_software_required
def test_repeat_pickle():

    myTR_O = repeat.Repeat(msa = TEST_MSA_O)

    test_pickle = os.path.join(path(), "test.pickle")
    myTR_O.write(test_pickle, 'pickle')
    myTR_O_new = repeat.Repeat.create(test_pickle, 'pickle')

    assert myTR_O.msa == myTR_O_new.msa
    assert myTR_O.sequence_type == myTR_O_new.sequence_type
    assert myTR_O.text == myTR_O_new.text

    if os.path.exists(test_pickle):
        os.remove(test_pickle)



# (C) 2015 Elke Schaper

"""
    :synopsis: Alignment of tandem repeat units.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import logging
import os
import subprocess
import tempfile
from Bio import AlignIO

log = logging.getLogger(__name__)

from tral.repeat import repeat
from tral import configuration

CONFIG_GENERAL = configuration.Configuration.instance().config
REPEAT_CONFIG = CONFIG_GENERAL["repeat"]

''' Some functions might overlap with repeat.gene_tree.align.'''


def realign_repeat(my_msa, aligner='mafft', sequence_type='AA', begin=None):

    # Create temporary working directory
    working_dir = tempfile.mkdtemp()
    log.debug("evolvedTR: Created temp directory: %s", working_dir)

    # Save my_TR to temp directory:
    msa_file = os.path.join(working_dir, 'msa_temp.faa')
    with open(msa_file, 'w') as msa_filehandle:
        for i, iMSA in enumerate(my_msa):
            msa_filehandle.write('> {0}\n{1}\n'.format(i, iMSA))

    if aligner == 'mafft':
        # Run Mafft
        # See http://mafft.cbrc.jp/alignment/software/manual/manual.html for choice of options.
        # The mafft result is in stdout. Check: Do you need to capture or
        # redirect the stderr?
        p = subprocess.Popen([REPEAT_CONFIG['ginsi'],
                              "--anysymbol",
                              "--quiet",
                              msa_file],
                             stdout=subprocess.PIPE)
        mafft_output = [line.decode('utf8').rstrip() for line in p.stdout]
        msa = []
        for iLine in mafft_output:
            if iLine[0] == '>':
                msa.append('')
            else:
                msa[-1] += iLine
        msa = [i for i in msa if i != '']
        log.debug('\n'.join(msa))
        p.wait()
        try:
            return msa
        except:
            error_note = (
                "Mafft could not successfully run the realignment for: "
                "\n".join(my_msa))
            logging.error(error_note)
            return None

    else:
        raise ValueError(
            'Currently, the aligner {} is not implemented.'.format(aligner))



import os
import pytest

from tral.hmm.hmm import HMM
from tral.hmm.hmm_viterbi import *
from tral.repeat import repeat
from tral.repeat import repeat_align

notfixed = pytest.mark.notfixed

@pytest.mark.no_external_software_required
def test_hmm_path_to_non_aligned_tandem_repeat_units():

    TEST = {"Double": ["AAAA", ["M1","M2","M1","M2"], 2, ["AA","AA"]],
            "Double_converted": ["ABCD", ["M2","M1","M2","M1"], 2, ["AB","CD"]],
            "Double_converted_complex": ["NNAAAACC", ["N","N","M2","M1","M2","M1","C","C"], 2, ["AA","AA"]],
            "Double_converted_complex_insertions": ["NNAIAAACC", ["N","N","M2","I2","M1","M2","M1","C","C"], 2, ["AIA","AA"]],
            "Single": ["AAAA", ["M1","M1","M1","M1"], 1, ["A","A","A","A"]],
            "Single_Complex": ["NNAAIIIAA", ["N","N","M1","M1","I1","I1","I1","M1","M1"], 1, ["A","AIII","A","A"]],
            "Long": ["GYRADKLADKLADKL", ["N","N","N","M1","M2","M3","M4","M1","M2","M3","M4","M1","M2","M3","M4"], 4, ["ADKL","ADKL","ADKL"]],
            }
    for test, p in TEST.items():
        test_repeat_msa = hmm_path_to_non_aligned_tandem_repeat_units(sequence = p[0], path = p[1], l_effective = p[2])
        assert test_repeat_msa == p[3]


def test_viterbi():

    # {Test_name: [Original_TR_MSA, Sequence, Viterbi_path, Refined_TR_MSA], ... }
    TEST = {"Single": [["A","A","A"], "AAAAAA", ["M1","M1","M1","M1","M1","M1"], ["A","A","A","A","A","A"]],
        "Double": [["AA","AA"], "AAAAAA", ["M1","M2","M1","M2","M1","M2"], ["AA","AA","AA"]],
        "Long": [["ADKL","ADKL"], "GYRADKLADKLADKL", ["N","N","N","M1","M2","M3","M4","M1","M2","M3","M4","M1","M2","M3","M4"], ["ADKL","ADKL","ADKL"]]
        }

    for test, p in TEST.items():
        test_repeat = repeat.Repeat(msa = p[0])
        test_hmm = HMM.create(input_format = "repeat", repeat = test_repeat)

        for iHMM in [test_hmm]:
            # Detect TRs on self.seq with hmm using the Viterbi algorithm.
            most_likely_path = iHMM.viterbi(p[1])
            assert type(most_likely_path) == list
            assert most_likely_path == p[2]

            unaligned_msa = hmm_path_to_non_aligned_tandem_repeat_units(p[1], most_likely_path, iHMM.l_effective)
            assert unaligned_msa == p[3]

            aligned_msa = repeat_align.realign_repeat(unaligned_msa)
            assert aligned_msa == p[3]

# The following test functions were copy-pasted from hmm_viterbi.py. They need to be either
# strongly adapted, or - if not useful - erased.
@notfixed
@pytest.mark.no_external_software_required
def test_conversion_multiple():
    """ Test <hmm_path_to_maximal_complete_tandem_repeat_units>.
        However, no positive set is defined right now.
    """
    assert 0, "Test not fixed"
    l_effective = 3
    lPaths = [['N','N','N', 'M0','M1','M2','M0','I1','M1','M2','M0','M1','M2','N','N','N'],
            ['N','N','N','M1','M2','M0','M1','M2','M0','M1','M2', 'M0', 'N','N','N'],
            ['N','N','N','M2','M0','M1','M2','M0','M1','M2','M0','M1', 'N','N','N']]
    lSequence = ['XXXABCADBCABCXXX','XXXBCABCABCAXXX','XXXCABCABCABXXX']
    lMSA = hmm_path_to_maximal_complete_tandem_repeat_units(lSequence, lPaths, 3)

@notfixed
@pytest.mark.no_external_software_required
def test_create_viterbi():
    """ This test needs to be fixed, the HMM cannot be initialised in this way.
    """
    assert 0, "Test not fixed"
    my_hmm = HMM()
    my_hmm.states = ['H', 'F']

    # Initialisation
    my_hmm.p_0 = {"H": 0.6, "F": 0.4}

    # Feed Values to p_t
    #my_hmm.p_t["START"] = {"H": 0.6, "F": 0.4}
    my_hmm.p_t["H"] = {"H": 0.7, "F": 0.3}
    my_hmm.p_t["F"] = {"H": 0.4, "F": 0.6}

    # emissions
    my_hmm.emissions = ["D", "C", "N"]

    # emission probabilities
    my_hmm.p_e['H'] = {"D":0.1, "C":0.4, "N":0.5}
    my_hmm.p_e['F'] = {"D":0.6, "C":0.3, "N":0.1}

    my_viterbi = Viterbi(my_hmm, "NCD")
    if my_viterbi.viterbi() == "HHF":
        print("Test SUCCESSFUL")
    else:
        print("Test FAILED")

@notfixed
def test_init_with_repeat():
    """ This test needs to be fixed.
    """
    assert 0, "Test not fixed"
    my_TR = Repeat(begin = 0, msa = ['A-G', 'ACG', 'ACG'], sequence_type = 'DNA')
    my_hmm = HMM.create(repeat=my_TR)
    from . import sequence
    my_sequence = sequence.Sequence()
    my_viterbi = Viterbi(my_hmm, my_sequence.sequence)
    v = my_viterbi.viterbi()
    "".join(v)



import pytest

from tral.repeat import repeat_align

TEST_MSA = ["MGKGYL---------------------------------------ALCSYNCKEA-INILSHLPSHHYN","TG--------------------------------------------------------------WVLCQ","IGRAYF---------------------------------------ELSEYMQAER-IFSEVRRIENYRV","EGMEIYSTTLWHLQK------------------------------DVALSVLSKDLTDMDKNSPEAWCA","AGNCFS---------LQREH-------------------------DIAIKFFQRA-IQVDPNYAYAYTL","LGHEFV--------------LTEEL--------------------DKALACFRNA-IRVNPRHYNAWYG","LGMIYY-------------------KQEKF---------------SLAEMHFQKA-LDINPQSSVLLCH","IGVVQH------------------------ALKKS----------EKALDTLNKA-IVIDPKNPLCKFH","RASVLF-----------------------------ANEKY-----KSALQELEEL-KQIVPKESLVYFL","IGKVYK----------------------------------KLGQTHLALMNFSWA-MDLDPKGAN----"]
TEST_REALIGNED_MSA = ['MGKGYLALC----SYNC-----KEAI-NILSHLPSHHYN', 'TGWVL--------------------------------CQ', 'IGRAYFELS----EYMQAERIFSEVR-RIEN-----YRV', 'EGMEIYSTTLWHLQKDVALSVLSKDLTDMDKNSPEAWCA', 'AGNCFSLQR----EHDIAIKFFQRAI-QVDPNYAYAYTL', 'LGHEFVLTE----ELDKALACFRNAI-RVNPRHYNAWYG', 'LGMIYYKQE----KFSLAEMHFQKAL-DINPQSSVLLCH', 'IGVVQHALK----KSEKALDTLNKAI-VIDPKNPLCKFH', 'RASVLFANE----KYKSALQELEELK-QIVPKESLVYFL', 'IGKVYKKLG----QTHLALMNFSWAM-DLDPKGA----N']


def test_repeat_alignment():

    ''' Test a realignment using Mafft's ginsi.'''

    REALIGNED_MSA = repeat_align.realign_repeat(my_msa = TEST_MSA)

    assert isinstance(REALIGNED_MSA, list)

    # Attention: If ginsi changes, the alignment will also change...
    assert REALIGNED_MSA == TEST_REALIGNED_MSA


#!python

import argparse
import csv
import datetime
import logging
import logging.config
import os
import pickle
import sys

from pyfaidx import Fasta

from tral import configuration
from tral.paths import config_file
from tral.repeat_list import repeat_list
from tral.sequence import sequence
from tral.hmm import hmm
from tral.hmm import hmm_io

logging.config.fileConfig(config_file("logging.ini"))
LOG = logging.getLogger('root')

CONFIG = configuration.Configuration.instance().config

# Shift REPEAT_LIST_TAG to configuration file and rename?
REPEAT_LIST_TAG = "all"
DE_NOVO_ALL_TAG = "denovo_all"
PFAM_ALL_TAG = "pfam_all"
DE_NOVO_TAG = "denovo"
PFAM_TAG = "pfam"
DE_NOVO_REFINED_TAG = "denovo_refined"
DE_NOVO_FINAL_TAG = "denovo_final"
FINAL_TAG = "final"


def workflow(
        sequences_file,
        hmm_annotation_file,
        hmm_dir,
        result_file,
        result_file_serialized,
        format,
        max_time,
        time_interval=3600,
        next_time=3600,
        **kwargs):
    ''' Annotate sequences with TRs from multiple sources, test and refine annotations.

     Save the annotations in a pickle.

     Args:
         sequences_file (str): Path to the pickle file containing a list of ``Sequence``
            instances.
         hmm_dir (str): Path to directory where all HMMs are stored as .pickles
         result_file (str): Path to the result file.
         max_time (str): Max run time in seconds

     Raises:
        Exception: If the pickle ``sequences_file`` cannot be loaded
        Exception: if the hmm_dir does not exist

    '''

    start = datetime.datetime.now()
    max_time, time_interval, next_time = int(
        max_time), int(time_interval), int(next_time)

    try:
        l_sequence = Fasta(sequences_file)
    except:
        raise Exception(
            "Cannot load putative pickle file sequences_file: {}".format(sequences_file))

    if not os.path.isdir(hmm_dir):
        try:
            os.makedirs(hmm_dir)
        except:
            raise Exception(
                "hmm_dir does not exists and could not be created: {}".format(hmm_dir))

    try:
        with open(hmm_annotation_file, 'rb') as fh:
            dHMM_annotation = pickle.load(fh)
    except:
        raise Exception(
            "Cannot load hmm_annotation_file: {}".format(hmm_annotation_file))

    basic_filter = CONFIG['filter']['basic']['dict']
    basic_filter_tag = CONFIG['filter']['basic']['tag']

    # Load previous results
    try:
        if not os.path.isdir(os.path.dirname(result_file)):
            os.makedirs(os.path.dirname(result_file))
    except:
        raise Exception(
            "Could not create path to result_file directory: {}".format(
                os.path.dirname(result_file)))

    try:
        with open(result_file, 'rb') as fh:
            dResults = pickle.load(fh)
    except:
        LOG.debug(
            "Could not load previous results file - perhaps non existant: {}".format(result_file))
        dResults = {}

    dHMM = {}
    for iS_pyfaidx in l_sequence:

        # If sequence is already included in results: continue.
        if iS_pyfaidx.name in dResults:
            continue

        elapsed_time = (datetime.datetime.now() - start).seconds
        if elapsed_time > max_time or elapsed_time > next_time:
            with open(result_file, 'wb') as fh:
                pickle.dump(dResults, fh)
            next_time = next_time + time_interval

        iS = sequence.Sequence(seq=str(iS_pyfaidx), name=iS_pyfaidx.name)

        LOG.debug("Work on sequence {}".format(iS))
        # 1. annotate_de_novo()
        denovo_repeat_list = iS.detect(
            denovo=True,
            repeat={
                "calc_pvalue": True})
        LOG.debug(denovo_repeat_list.repeats)
        for iTR in denovo_repeat_list.repeats:
            iTR.model = None

        # 2. annotate_TRs_from_hmmer()
        if iS.name in dHMM_annotation:
            lHMM = dHMM_annotation[iS.name]
            infoNRuns = len(lHMM)
            LOG.debug(
                "{} Viterbi runs need to be performed.".format(infoNRuns))
            lHMM = set(lHMM)
            infoNHMM = len(lHMM)
            LOG.debug(
                "These derive from {} independent HMMs.".format(infoNHMM))
            # Load all HMM pickles needed for the particular sequence.
            for hmm_ID in lHMM:
                if hmm_ID not in dHMM:
                    dHMM[hmm_ID] = hmm.HMM.create(
                        file_format="pickle",
                        file=os.path.join(
                            hmm_dir,
                            hmm_ID +
                            ".pickle"))

            pfam_repeat_list = iS.detect(
                [dHMM[hmm_ID] for hmm_ID in lHMM], repeat={"calc_pvalue": True})
            for iTR, hmm_ID in zip(pfam_repeat_list.repeats, lHMM):
                iTR.model = hmm_ID
                iTR.TRD = "PFAM"
        else:
            pfam_repeat_list = None

        # 3. merge_and_basic_filter()
        all_repeat_list = denovo_repeat_list + pfam_repeat_list
        iS.set_repeatlist(all_repeat_list, REPEAT_LIST_TAG)
        iS.set_repeatlist(denovo_repeat_list, DE_NOVO_ALL_TAG)
        iS.set_repeatlist(pfam_repeat_list, PFAM_ALL_TAG)

        rl_tmp = iS.get_repeatlist(REPEAT_LIST_TAG)
        if iS.get_repeatlist(REPEAT_LIST_TAG):
            for iB in basic_filter.values():
                rl_tmp = rl_tmp.filter(**iB)
        else:
            rl_tmp = iS.get_repeatlist(REPEAT_LIST_TAG)
        iS.set_repeatlist(rl_tmp, basic_filter_tag)
        iS.set_repeatlist(
            rl_tmp.intersection(
                iS.get_repeatlist(PFAM_ALL_TAG)), PFAM_TAG)
        iS.set_repeatlist(
            rl_tmp.intersection(iS.get_repeatlist(DE_NOVO_ALL_TAG)),DE_NOVO_TAG)

        # 4. calculate_overlap()

        # Perform common ancestry overlap filter and keep PFAMs
        criterion_pfam_fixed = {
            "func_name": "none_overlapping_fixed_repeats",
            "rl_fixed": iS.get_repeatlist(PFAM_TAG),
            "overlap_type": "common_ancestry"}

        iS.d_repeatlist[DE_NOVO_TAG] = iS.get_repeatlist(DE_NOVO_TAG).filter(**criterion_pfam_fixed)

        # Choose only the most convincing de novo TRs
        criterion_filter_order = {
            "func_name": "none_overlapping", "overlap": (
                "common_ancestry", None), "l_criterion": [
                ("pvalue", "phylo_gap01"), ("divergence", "phylo_gap01")]}
        iS.d_repeatlist[DE_NOVO_TAG] = iS.get_repeatlist(DE_NOVO_TAG).filter(**criterion_filter_order)

        # 5. refine_denovo()
        denovo_final = []
        denovo_refined = [None] * len(iS.get_repeatlist(DE_NOVO_ALL_TAG).repeats)
        for i, iTR in enumerate(iS.get_repeatlist(DE_NOVO_ALL_TAG).repeats):
            if not iTR in iS.get_repeatlist(DE_NOVO_TAG).repeats:
                continue
            # Create HMM from TR
            denovo_hmm = hmm.HMM.create(file_format='repeat', repeat=iTR)
            # Run HMM on sequence
            denovo_refined_rl = iS.detect(lHMM=[denovo_hmm])
            append_refined = False
            if denovo_refined_rl and denovo_refined_rl.repeats:
                iTR_refined = denovo_refined_rl.repeats[0]
                iTR_refined.TRD = iTR.TRD
                iTR_refined.model = "cpHMM"
                denovo_refined[i] = iTR_refined
                # Check whether new and old TR overlap. Check whether new TR is
                # significant. If not both, put unrefined TR into final.
                if repeat_list.two_repeats_overlap(
                        "shared_char",
                        iTR,
                        iTR_refined):
                    rl_tmp = repeat_list.RepeatList([iTR_refined])
                    LOG.debug(iTR_refined.msa)
                    for iB in basic_filter.values():
                        rl_tmp = rl_tmp.filter(**iB)
                    if rl_tmp.repeats:
                        append_refined = True
            else:
                denovo_refined[i] = False
            if append_refined:
                denovo_final.append(iTR_refined)
            else:
                denovo_final.append(iTR)

        iS.set_repeatlist(
            repeat_list.RepeatList(denovo_refined),
            DE_NOVO_REFINED_TAG)
        iS.set_repeatlist(
            repeat_list.RepeatList(denovo_final),
            DE_NOVO_FINAL_TAG)
        iS.set_repeatlist(
            iS.get_repeatlist(DE_NOVO_FINAL_TAG) +
            iS.get_repeatlist(PFAM_TAG),
            FINAL_TAG)

        dResults[iS.name] = iS

    # 6.a Save results as pickle
    with open(result_file, 'wb') as fh:
        pickle.dump(dResults, fh)

    # 6.b Save serialized results
    with open(result_file_serialized, 'w') as fh_o:

        if format == 'tsv':
            header = [
                "ID",
                "MSA",
                "begin",
                "pvalue",
                "l_effective",
                "n",
                "n_effective",
                "TRD",
                "model"]
        fh_o.write("\t".join(header))

        for iS in dResults.values():
            for iTR in iS.get_repeatlist(FINAL_TAG).repeats:
                if format == 'tsv':
                    try:
                        data = [
                            str(i) for i in [
                                iS.name,
                                " ".join(
                                    iTR.msa),
                                iTR.begin,
                                iTR.pvalue("phylo_gap01"),
                                iTR.l_effective,
                                iTR.n,
                                iTR.n_effective,
                                iTR.TRD,
                                iTR.model]]
                    except:
                        print(iTR)
                        raise Exception(
                            "(Could not save data for the above TR.)")

                fh_o.write("\n" + "\t".join(data))

    print("DONE")


def file_preparation(
        annotation_data_file,
        annotation_data_output,
        hmm_raw_file,
        hmm_dir):

    create_hmm_files(hmm_raw_file, hmm_dir)

    read_pfam_uniprot(annotation_data_file, annotation_data_output)


def create_hmm_files(hmm_file, output_file):

    for hmmer_probabilities in hmm_io.read(hmm_file):
        iID = hmmer_probabilities['id'].split(".")[0]
        iHMM = hmm.HMM(hmmer_probabilities)
        file = os.path.join(output_file, "{}.pickle".format(iID))
        iHMM.write(file, "pickle")


def read_pfam_uniprot(annotation_data_file, output_file):
    ''' annotation_file from:
        http://www.uniprot.org/uniprot/?query=database:(type:pfam%20AND%20*)&fil=&sort=score '''

    p = {}
    if annotation_data_file:
        try:
            with open(annotation_data_file) as f:
                reader = csv.reader(f, delimiter="\t")
                for row in reader:
                    p[row[0]] = row[1][:-1].split(";")
        except:
            raise Exception("Cannot load sequence annotation file annotation_data_file: {}".format(
                annotation_data_file))

    with open(output_file, 'wb') as fh:
        pickle.dump(p, fh)


def main():

    pars = read_commandline_arguments()

    # Update configuration
    if "sequence_type" in pars:
        CONFIG["sequence_type"] = pars["sequence_type"]
    if "detectors" in pars:
        CONFIG["sequence"]["repeat_detection"][
            CONFIG["sequence_type"]] = pars["detectors"]
    if "significance_test" in pars:
        CONFIG["repeat_score"]["score_calibration"] = pars["significance_test"]

    if pars["method"] == "file_preparation":
        file_preparation(
            pars["hmm_annotation_raw"],
            pars["hmm_annotation"],
            pars["hmm_raw"],
            pars["hmm"])
    elif pars["method"] == "workflow":
        workflow(
            pars["input"],
            pars["hmm_annotation"],
            pars["hmm"],
            pars["output"],
            pars["output_serialized"],
            pars["format"],
            pars["time"])


def read_commandline_arguments():
    parser = argparse.ArgumentParser(
        description='Process tandem repeat detection options')
    parser.add_argument('method', metavar='method_name', type=str,
                        help='The name of the method to be executed.')
    parser.add_argument('-i', '--input', type=str,
                        help='The path to the input file.')
    parser.add_argument('-o', '--output', type=str,
                        help='The path to the output file.')
    parser.add_argument('-os', '--output_serialized',
                        help='The path to the serialized output file.')
    parser.add_argument('-f', '--format', type=str,
                        help='The serialization format, e.g. tsv')
    parser.add_argument('-t', '--time', type=str,
                        help='The maximum runtime')
    parser.add_argument('-hmm', '--hmm', type=str,
                        help='The path to the dir with HMM pickles')
    parser.add_argument('-hmm_raw', '--hmm_raw', type=str,
                        help='The path to the raw file with HMM models')
    parser.add_argument(
        '-hmm_annotation',
        '--hmm_annotation',
        type=str,
        help='The path the pickle with sequence to HMM mappings')
    parser.add_argument(
        '-hmm_annotation_raw',
        '--hmm_annotation_raw',
        type=str,
        help='The path the raw file with sequence to HMM mappings')

    pars = vars(parser.parse_args())
    pars = {key: value for key, value in pars.items() if value is not None}
    return pars


if __name__ == "__main__":
    main()



# (C) 2015 Elke Schaper

"""
    :synopsis: Input/output for tandem repeats.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import Bio.Seq
import random
import logging
import itertools
import os
import re
import shutil
import tempfile
import subprocess
import numpy as np

from tral.paths import DATA_DIR, EXEC_DIR

LOG = logging.getLogger(__name__)


###################### SAVE REPEAT #######################################

def save_repeat_fasta(tandem_repeats, file):
    """ save multiple <tandem_repeats> in Fasta format in specified <file>

        At current, only one TR per sequence can be defined, as the identifiers
        in the dict <tandem_repeats> must be unique.

        Parameters: Dict of tandem repeats and identifiers.
            e.g. {'ENSP00012': msa1, 'ENSP00013': msa2}

            >ID
            GHKI
            GHKI
            GH--
    """

    with open(file, 'w', newline='\n') as f:
        for identifier, msa in tandem_repeats.items():
            f.write(">{0}\n".format(identifier))
            f.write("\n".join(msa) + "\n\n")


def save_repeat_stockholm(tandem_repeat, file):
    """ save <tandem_repeat> in STOCKHOLM format in specified <file>

        Parameters: Tandem repeat MSA.
            e.g. ["ACDEF-", "ACCDEF"]

            More information in
            ftp://selab.janelia.org/pub/software/hmmer3/3.0/Userguide.pdf
            STOCKHOLM Format Example:

            # STOCKHOLM 1.0
            seq1 ACDEF...GHIKL
            seq2 ACDEF...GHIKL
            seq3 ...EFMNRGHIKL

            seq1 MNPQTVWY
            seq2 MNPQTVWY
            seq3 MNPQT...
            //

    """

    with open(file, 'w', newline='\n') as fh:
        fh.write("# STOCKHOLM 1.0\n")
        for i, iMSA in enumerate(tandem_repeat):
            fh.write("{} {}\n".format(str(i), iMSA))
        fh.write("//")


def save_repeat_treks(tandem_repeats, file):
    """ Save multiple `tandem_repeats` in T-REKS format in specified `file`

        At current, only one TR per sequence can be defined, as the identifiers
        in the dict <tandem_repeats> must be unique.

        Parameters: Dict of tandem repeats and identifiers.
            e.g. {'ENSP00012': [msa1, begin1], 'ENSP00013': [msa2, begin2]}

            T-REKS format example:

            >a
            Length: 3 residues - nb: 3  from  1 to 10 - Psim:0.8076923076923077 region Length:42
            GHKI
            GHKI
            GH--
            **********************

            Length: 3 residues - nb: 2  from  21 to 27 - Psim:0.7857142857142857 region Length:22
            GHKI
            GH--
            **********************

            >b
            Length: 3 residues - nb: 3  from  1 to 10 - Psim:0.8095238095238095 region Length:20
            UIR
            UIR
            UIR
    """

    with open(file, 'w', newline='\n') as fh:
        for identifier, info in tandem_repeats.items():
            fh.write(">{0}\n".format(identifier))
            fh.write("Length: from {0} to\n".format(str(info[1])))
            fh.write("\n".join(info[0]) + "\n" + "*" * 22 + "\n\n")


############################## READ REPEAT SEQUENCE ######################


def read_fasta(seq_filename, sequence_type='AA'):
    """ Read repeat from file in fasta format.

    Read repeat from file in fasta format.

    Args:
        seq_filename (str):  Path to the repeats containing fasta file
        sequence_type (str): Either "AA" or "DNA"

    Returns:
        (list of Repeat): A list of Repeat instances.
    """

    pat_start = re.compile(r">(.*)")
    pat_repeat_unit = re.compile(r"([\w\.\-]+)")

    # Our possible parser states:
    #
    # 1: searching for sequence name
    # 2: searching for repeat units

    repeats = {}
    state = 1
    with open(seq_filename, "rt") as infile:

        for i, line in enumerate(infile):
            LOG.debug("Line {0}: {1}".format(i, line[0:-1]))
            if 1 == state:
                match = pat_start.match(line)
                if match:
                    LOG.debug(" * (1->2) Found start")
                    LOG.debug("Start: %s", match.group(1))
                    name = match.group(1)
                    repeats[name] = []
                    state = 2

            elif 2 == state:
                match = pat_repeat_unit.match(line)
                if match:
                    LOG.debug(" * (2->2) Found Repeat unit")
                    LOG.debug("Repeat Unit: %s", match.group(1))
                    repeat_unit = match.group(1)
                    repeats[name].append(repeat_unit.replace(".", "-").upper())
                else:
                    LOG.debug(" * (2->1) Found NO further Repeat unit")
                    state = 1

    for i_name, iR in repeats.items():
        yield iR, sequence_type


################################### SIMULATE SEQUENCE ####################

def evolved_tandem_repeats(l, n, n_samples, sequence_type, job_id='job_id',
                           mutation_rate=50, tree='star',
                           indel_rate_per_site=False,
                           return_type='repeat'):
    """ Simulate evolved sequences with ALF.

    Simulate evolved sequences with ALF:
    Dalquen, D. A., Anisimova, M., Gonnet, G. H. & Dessimoz, C.
    ALF--a simulation framework for genome evolution. Molecular Biology and Evolution 29,
    1115–1123 (2012).

    Args:
        l (int): The length of the repeat unit
        n (int): The number of repeat units in the tandem repeat
        n_samples (int):  The number of samples
        sequence_type (str): Either "AA" or "DNA"
        job_id (str): A tag for files produces with ALF, and result files.
        mutation_rate (float): The mutation rate.
        tree (str): The type of tree, e.g. "star" or "birthdeath"
        indel_rate_per_site (int or False): The indel rate per site.
        return_type (str): Either "repeat" or "list"

        sequence_length (int): The total length of the simulated sequence

    Returns:
        Return type depends on ``return_type``: ``Repeat`` or ``Bio.Seq.Seq`` instance.
    """

    runfile_template = os.path.join(DATA_DIR, "ALF", "template.drw")
    alf_exec = os.path.join(EXEC_DIR, "alfsim")
    # create temporary directory
    working_dir = tempfile.mkdtemp()
    LOG.debug("evolvedTR: Created tempfile: %s", working_dir)

    # create working dir
    if not os.path.isdir(working_dir):
        os.makedirs(working_dir)

    # Copy template file and append job specific info
    runfilename = "alf.drw"
    shutil.copyfile(runfile_template, os.path.join(working_dir, runfilename))

    with open(os.path.join(working_dir, runfilename), "a") as runfile:
        if indel_rate_per_site:
            # insertion rate per site and PAM. E.g. for PAM=40 expect
            # aaGainRate*40 insertions per site.
            runfile.write(
                "aaGainRate := " + str(indel_rate_per_site / mutation_rate) + ";\n")
            runfile.write(
                "aaLossRate := " + str(indel_rate_per_site / mutation_rate) + ";\n")
            runfile.write("maxIndelLength := 50;\n")
            runfile.write("indelModel := 'ZIPF';\n")
            runfile.write("Z_c := 1.821;\n")
            runfile.write("DawgPlacement := true;\n")
        runfile.write("uuid := '" + job_id + "';\n")
        runfile.write("mname := '" + job_id + "';\n")
        runfile.write("protStart := " + str(n_samples) + ";\n")
        runfile.write("NSpecies := " + str(n) + ";\n")
        runfile.write("minGeneLength := " + str(l) + ";\n")
        runfile.write("mutRate := " + str(mutation_rate) + ";\n")
        runfile.write("wdir := '" + working_dir + "';\n")

        tree_length = n * mutation_rate
        # parameters concerning the species tree
        if tree == 'star':
            # BDTree, ToLSample, Custom
            runfile.write("treeType := 'Custom':\n")
            runfile.write(
                "tree := MakeStarTree(BirthDeathTree(0.1,0.1," + str(n) + ",10),mutRate):\n")
            runfile.write("treeLength:= %d ;\n" % tree_length)
        else:
            # BDTree, ToLSample, Custom
            runfile.write("treeType := 'BDTree':\n")
            # From the last discussion with Daniel, only the ration of birth to
            # death rate matters, if we scale tree to match pam distance
            runfile.write("birthRate := 0.01:\n")
            runfile.write("deathRate := 0.01:\n")
            # for BDTree: should resulting tree be ultrametric, e.g. all leaves
            # have same distance to origin?
            runfile.write("ultrametric := false:\n")
            runfile.write("treeLength:= %d ;\n" % tree_length)
            # DANIEL: Is a tree := missing?
        if tree not in {'star', 'birthdeath'}:
            LOG.warning(
                "evolvedTR: tree input %s not known, assuming birthdeath tree",
                tree)

        # parameters concerning the substitution models
        if sequence_type == 'AA':
            runfile.write(
                "substModels := [SubstitutionModel('CustomP', ['" +
                os.path.join(
                    DATA_DIR,
                    'ALF',
                    'lg.dat') +
                "'])];\n")
            # CHECK! DOES
            # substModels := [SubstitutionModel('LG')];
            # WORK?
        elif sequence_type == 'DNA':
            runfile.write(
                "substModels := [SubstitutionModel('TN93', [.3, .4, .7],[seq(0.25,4)], true)]:\n")
            # CHECK! DO THE EMPIRICAL PARAMETERS MAKE SENSE AT ALL?
            # compare to http://people.inf.ethz.ch/ddalquen/alf/ALF_manual.pdf
        else:  # CODONS
            runfile.write("substModels := [SubstitutionModel('CPAM')];\n")

    # Determine ALF MSA output file path
    if sequence_type == 'AA':
        infile = os.path.join(working_dir, job_id, "MSA", "MSA_all_aa.phy")
    elif sequence_type == 'DNA':
        infile = os.path.join(working_dir, job_id, "MSA", "MSA_all_dna.phy")
    else:  # CODONS
        infile = os.path.join(
            working_dir,
            job_id,
            "MSA",
            "MSA_all_codon.phy")

    for i in range(10):
        with open(os.path.join(working_dir, "out.txt"), "w") as outfile:
            alf_process = subprocess.Popen([alf_exec,
                                            runfilename],
                                           cwd=working_dir,
                                           stdout=outfile,
                                           stderr=outfile,
                                           close_fds=True)
            alf_process.wait()
        if os.path.isfile(infile):
            break
    else:
        LOG.error('ALFSIM was not able to produce simulated sequence.')

    # shutil.rmtree('/cluster/home/infk/eschaper/spielwiese/')
    #shutil.copytree(working_dir, '/cluster/home/infk/eschaper/spielwiese/')

    """ Read in MSA data from ALF
        The following is a short parser for ALFsim output files
        yielding MSAs as lists of strings,
        based on a special flavour of  Felstein stein MSA files """

    # find a repeat uni
    pattern_start = re.compile("\d+ \d+")
    pattern_seq = re.compile("\S+[ ]+([A-Z\-]+)")

    # Our possible parser states:
    #
    # state 1: Find beginning of MSA (Felsenstein)
    # state 2: Find all repeat units

    # protein ::=
    #    \d \d
    #    \S/\S    repeatunits
    #

    state = 1
    with open(infile, "r") as infile:
        for i, line in enumerate(infile):
            LOG.debug("Line %d: %s", i, line[0:-1])

            if 1 == state:  # Find first repeat unit & save begin
                search = pattern_start.search(line)
                if search:
                    LOG.debug(" *(1->2) Found MSA start")
                    state = 2
                    msa = []

            elif 2 == state:  # Find all repeat units
                search = pattern_seq.search(line)
                if search:
                    LOG.debug(" *(2->2) Found another repeat unit")
                    msa.append(search.group(1))
                else:
                    LOG.debug(" *(2->1) repeat region finished, yielding.")
                    state = 1
                    # YIELD IF WE HAVE FOUND AT LEAST TWO REPEAT UNITS:
                    if len(msa) > 1:
                        if return_type == 'repeat':
                            yield repeat.Repeat(begin=0, msa=msa, sequence_type=sequence_type)
                        elif return_type == 'list':
                            yield msa
                        else:
                            LOG.debug("YIELD: %s",
                                      "".join(msa).replace('-', ''))
                            yield Bio.Seq.Seq("".join(msa).replace('-', ''), sequence_type)

    # delete temporary directory
    try:
        shutil.rmtree(working_dir)
    except OSError:
        pass


def random_sequence(n_samples, sequence_type='AA', return_type='repeat',
                    equilibrium_frequencies='human', l=0, n=0,
                    sequence_length=0):
    """ Simulate random sequence locally.

    Simulate random sequence locally.

    Args:
        n_samples (int):  The number of samples
        sequence_type (str): Either "AA" or "DNA"
        return_type (str): Either "repeat" or "list"
        equilibrium_frequencies (str): Only "human" option available at current
        l (int): The length of the repeat unit
        n (int): The number of repeat units in the tandem repeat
        sequence_length (int): The total length of the simulated sequence

    Returns:
        Return type depends on ``return_type``.
    """

    if sequence_length == 0 and (l == 0 or n == 0):
        LOG.error('The specified sequence_length or the product of l and n was'
                  ' set to 0 for random_sequence simulation')
    else:
        if sequence_length == 0:
            sequence_length = l * n

        if equilibrium_frequencies == 'human':
            file = os.path.join(DATA_DIR, 'Random', "_".join(
                [sequence_type, equilibrium_frequencies, '3']) + '.txt')

        with open(file, 'r') as f:
            a = f.readline()[:-1]
        b = [i.split(' ') for i in a.split('  ')]
        frequencies = {i[0]: int(i[1]) for i in b}
        alphabet = np.unique(i[0] for i in frequencies.keys())

        for _ in range(n_samples):
            seed_int = random.randint(1, sum(frequencies.values()))
            for key, value in frequencies.items():
                seed_int -= value
                if seed_int <= 0:
                    seed = key
                    break

            dimer = [''.join(i) for i in itertools.product(alphabet, repeat=2)]
            third_letter_frequencies = {
                iD: {
                    iA: frequencies[
                        iD +
                        iA] for iA in alphabet} for iD in dimer}

            sequence = seed
            for _ in range(sequence_length - 3):
                next_int = random.randint(
                    1, sum(third_letter_frequencies[sequence[-2:]].values()))
                for key, value in third_letter_frequencies[
                        sequence[-2:]].items():
                    next_int -= value
                    if next_int <= 0:
                        sequence += key
                        break

            # return Repeat instances
            if return_type == 'repeat' and not l == 0 and not n == 0:
                yield [sequence[i * l:(i + 1) * l] for i in range(n)], sequence_type
            elif return_type == 'list':
                yield sequence
            else:  # return seqIO instances
                yield Bio.Seq.Seq(sequence, sequence_type)



import os
import pytest

from tral.hmm.hmm import HMM
from tral.hmm import hmm_io
from tral.repeat.repeat import Repeat


NUM_STATES = 8
AMINOACIDS = list('ACDEFGHIKLMNPQRSTVWY')
NUM_AMINOACIDS = len(AMINOACIDS)
NUM_TRANSITIONS = 7

CARCINUSTATIN_ID = 'PF08261.7'
WRONG_CARCINUSTATIN_ID = 'PF05932'
SHORT_CARCINUSTATIN_ID = 'PF08261'

# HMM data dictionary key names
INSERTION_EMISSIONS = 'insertion_emissions'
EMISSIONS = 'emissions'
TRANSITIONS = 'transition'
COMPO_KEY_NAME = 'COMPO'
LETTER_KEY_NAME = 'letters'
ID_KEY_NAME = 'id'

# Test file names
TEST_FILE_WITH_ID = 'carcinustatin.hmm'
TEST_FILE_WITHOUT_ID = 'carcinustatin_no_id.hmm'


TEST_REPEAT_MSA_SINGLE = ["A","A","A"]

notfixed = pytest.mark.notfixed

@pytest.fixture
def path():
    """Return the path to the test data files.
    """
    return os.path.join(os.path.abspath('.'), 'hmm', 'test')

@pytest.mark.no_external_software_required
def test_single_hmm_with_id_read(path):
    test_dict_list = list(HMM.read(os.path.join(path, TEST_FILE_WITH_ID)))
    assert len(test_dict_list) == 1
    test_dict = test_dict_list[0]

    compare_carcinustatin(test_dict)

    assert test_dict[ID_KEY_NAME] == CARCINUSTATIN_ID

@pytest.mark.no_external_software_required
def test_single_hmm_without_id_read(path):
    test_dict_list = list(HMM.read(os.path.join(path, TEST_FILE_WITHOUT_ID)))
    assert len(test_dict_list) == 1
    test_dict = test_dict_list[0]

    compare_carcinustatin(test_dict)

    assert test_dict[ID_KEY_NAME] is None


@pytest.mark.no_external_software_required
def test_single_hmm_no_id_with_query_read(path):
    test_dict_list = list(HMM.read(os.path.join(path, TEST_FILE_WITHOUT_ID),
                              id=CARCINUSTATIN_ID))
    assert len(test_dict_list) == 0


@pytest.mark.no_external_software_required
def test_single_hmm_with_wrong_query_read(path):
    test_dict_list = list(HMM.read(os.path.join(path, TEST_FILE_WITH_ID),
                              id=WRONG_CARCINUSTATIN_ID))
    assert len(test_dict_list) == 0


@pytest.mark.no_external_software_required
def test_single_hmm_with_short_query_read(path):
    test_dict_list = list(hmm_io.read(os.path.join(path, TEST_FILE_WITH_ID),
                              id=SHORT_CARCINUSTATIN_ID))
    assert len(test_dict_list) == 1
    test_dict = test_dict_list[0]

    compare_carcinustatin(test_dict)

    assert test_dict[ID_KEY_NAME] == CARCINUSTATIN_ID

@pytest.mark.no_external_software_required
def compare_carcinustatin(test_dict):
    for i in range(0, NUM_STATES):
        assert str(i + 1) in test_dict.keys()
        assert INSERTION_EMISSIONS in test_dict[str(i + 1)].keys()
        assert EMISSIONS in test_dict[str(i + 1)].keys()
        assert TRANSITIONS in test_dict[str(i + 1)].keys()
        assert len(test_dict[str(i + 1)][INSERTION_EMISSIONS]) ==\
            NUM_AMINOACIDS
        assert len(test_dict[str(i + 1)][EMISSIONS]) ==\
            NUM_AMINOACIDS
        assert len(test_dict[str(i + 1)][TRANSITIONS]) == NUM_TRANSITIONS

    assert INSERTION_EMISSIONS in test_dict[COMPO_KEY_NAME].keys()
    assert EMISSIONS in test_dict[COMPO_KEY_NAME].keys()
    assert TRANSITIONS in test_dict[COMPO_KEY_NAME].keys()
    assert len(test_dict[COMPO_KEY_NAME][INSERTION_EMISSIONS]) == NUM_AMINOACIDS
    assert len(test_dict[COMPO_KEY_NAME][EMISSIONS]) == NUM_AMINOACIDS
    assert len(test_dict[COMPO_KEY_NAME][TRANSITIONS]) == NUM_TRANSITIONS

    assert len(set(test_dict[LETTER_KEY_NAME]) ^ set(AMINOACIDS)) == 0

    assert ID_KEY_NAME in test_dict.keys()



import os
import pytest

from tral.hmm.hmm import HMM
from tral.sequence import sequence
from tral.repeat import repeat
from tral.repeat_list import repeat_list


TEST_SEQUENCE = "FFAAAAAAFF"
# Zinc finger protein Q9BRR0
TEST_SEQUENCE_Q9BRR0 = "MARELSESTALDAQSTEDQMELLVIKVEEEEAGFPSSPDLGSEGSRERFRGFRYPEAAGPREALSRLRELCRQWLQPEMHSKEQILELLVLEQFLTILPGNLQSWVREQHPESGEEVVVLLEYLERQLDEPAPQVSGVDQGQELLCCKMALLTPAPGSQSSQFQLMKALLKHESVGSQPLQDRVLQVPVLAHGGCCREDKVVASRLTPESQGLLKVEDVALTLTPEWTQQDSSQGNLCRDEKQENHGSLVSLGDEKQTKSRDLPPAEELPEKEHGKISCHLREDIAQIPTCAEAGEQEGRLQRKQKNATGGRRHICHECGKSFAQSSGLSKHRRIHTGEKPYECEECGKAFIGSSALVIHQRVHTGEKPYECEECGKAFSHSSDLIKHQRTHTGEKPYECDDCGKTFSQSCSLLEHHRIHTGEKPYQCSMCGKAFRRSSHLLRHQRIHTGDKNVQEPEQGEAWKSRMESQLENVETPMSYKCNECERSFTQNTGLIEHQKIHTGEKPYQCNACGKGFTRISYLVQHQRSHVGKNILSQ"


TEST_REPEAT_MSA_SINGLE = ["A","A","A"]
TEST_RESULT_REPEAT_MSA_SINGLE = ["A","A","A","A","A","A"]

TEST_REPEAT_MSA_DOUBLE = ["AA","AA"]
TEST_RESULT_REPEAT_MSA_DOUBLE = ["AA","AA","AA"]

TEST_RESULT_REPEAT_MSA_LONG = [10*"A", 10*"A"]
TEST_SEQUENCE_A = 18*"A"

TEST_RESULT_REPEAT_MSA_SUPER_LONG = [100*"A", 100*"A"]
TEST_SEQUENCE_SUPER_LONG_A = 200*"A"


#Test file names
TEST_FILE_WITH_ID = 'carcinustatin.hmm'

TEST_SEQUENCE_TAG = 'test_sequence_tag'


notfixed = pytest.mark.notfixed

@pytest.fixture
def path():
    """Return the path to the test data files.
    """
    return os.path.join(os.path.abspath('.'), 'hmm', 'test')


@pytest.mark.no_external_software_required
def test_initialise_sequence():
    test_seq = sequence.Sequence(TEST_SEQUENCE)
    assert test_seq.seq == TEST_SEQUENCE


@pytest.mark.no_external_software_required
def test_detect_repeats_with_hmm():
    test_hmm = HMM.create(input_format = 'hmmer', file = os.path.join(path(), TEST_FILE_WITH_ID))
    test_seq = sequence.Sequence(TEST_SEQUENCE)
    test_optimized_repeat = test_seq.detect([test_hmm])


#@notfixed
def test_detect_repeats_with_repeat():

    test_repeat = repeat.Repeat(msa = TEST_REPEAT_MSA_DOUBLE)
    test_hmm = HMM.create(input_format = 'repeat', repeat = test_repeat)
    test_seq = sequence.Sequence(TEST_SEQUENCE)
    test_optimized_repeat = test_seq.detect([test_hmm])
    assert type(test_optimized_repeat) == repeat_list.RepeatList
    assert len(test_optimized_repeat.repeats) == 1
    assert test_optimized_repeat.repeats[0].msa == TEST_RESULT_REPEAT_MSA_DOUBLE

    test_repeat = repeat.Repeat(msa = TEST_REPEAT_MSA_SINGLE)
    test_hmm = HMM.create(input_format = 'repeat', repeat = test_repeat)
    test_optimized_repeat = test_seq.detect([test_hmm])
    assert type(test_optimized_repeat) == repeat_list.RepeatList
    assert len(test_optimized_repeat.repeats) == 1
    assert test_optimized_repeat.repeats[0].msa == TEST_RESULT_REPEAT_MSA_SINGLE


@pytest.mark.no_external_software_required
def test_too_big_hmms():

    test_repeat = repeat.Repeat(msa = TEST_RESULT_REPEAT_MSA_LONG)
    test_hmm = HMM.create(input_format = 'repeat', repeat = test_repeat)
    test_seq = sequence.Sequence(TEST_SEQUENCE_A)
    test_optimized_repeat = test_seq.detect([test_hmm])
    assert type(test_optimized_repeat) == repeat_list.RepeatList
    assert len(test_optimized_repeat.repeats) == 0

    test_repeat = repeat.Repeat(msa = TEST_RESULT_REPEAT_MSA_SUPER_LONG)
    test_hmm = HMM.create(input_format = 'repeat', repeat = test_repeat)
    test_seq = sequence.Sequence(TEST_SEQUENCE_SUPER_LONG_A)
    test_optimized_repeat = test_seq.detect([test_hmm])
    assert type(test_optimized_repeat) == repeat_list.RepeatList
    assert len(test_optimized_repeat.repeats) == 0

#@notfixed
def test_detect_repeats_denovo():

    test_parameters = {"detection": {"detectors": ["TRUST"]}}

    test_seq = sequence.Sequence(TEST_SEQUENCE_Q9BRR0)
    test_optimized_repeat = test_seq.detect(denovo = True, **test_parameters)

    assert type(test_optimized_repeat) == repeat_list.RepeatList
    assert len(test_optimized_repeat.repeats) == 3


@pytest.mark.no_external_software_required
def test_sequence_pickle():

    test_seq = sequence.Sequence(TEST_SEQUENCE)

    test_pickle = os.path.join(path(), "test.pickle")
    test_seq.write(test_pickle, 'pickle')
    test_seq_new = sequence.Sequence.create(test_pickle, 'pickle')

    assert test_seq.seq == test_seq_new.seq

    test_repeat = repeat.Repeat(msa = TEST_REPEAT_MSA_DOUBLE)
    test_hmm = HMM.create(input_format = 'repeat', repeat = test_repeat)
    test_optimized_repeat = test_seq.detect([test_hmm])
    test_seq.set_repeatlist(test_optimized_repeat, TEST_SEQUENCE_TAG)

    assert type(test_optimized_repeat) == repeat_list.RepeatList
    assert list(test_seq.d_repeatlist.keys()) == [TEST_SEQUENCE_TAG]
    assert type(test_seq.d_repeatlist[TEST_SEQUENCE_TAG]) == repeat_list.RepeatList
    assert test_seq.d_repeatlist[TEST_SEQUENCE_TAG].repeats

    test_retrieved_repeatlist = test_seq.get_repeatlist(TEST_SEQUENCE_TAG)
    assert test_retrieved_repeatlist == test_optimized_repeat

    test_seq.write(test_pickle, 'pickle')
    test_seq_new = sequence.Sequence.create(test_pickle, 'pickle')

    assert test_seq.d_repeatlist.keys() == test_seq_new.d_repeatlist.keys()
    assert test_seq.d_repeatlist[TEST_SEQUENCE_TAG].repeats[0].msa == test_seq_new.d_repeatlist[TEST_SEQUENCE_TAG].repeats[0].msa

    if os.path.exists(test_pickle):
        os.remove(test_pickle)



# (C) 2011, Alexander Korsunsky
# (C) 2011-2015 Elke Schaper

"""
    :synopsis: The Repeat Class.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

from collections import defaultdict
from copy import deepcopy
import logging
import numpy as np
import pickle
import re

from tral.repeat import repeat_score, repeat_pvalue, repeat_io
from tral import configuration

LOG = logging.getLogger(__name__)

CONFIG_GENERAL = configuration.Configuration.instance().config
CONFIG = CONFIG_GENERAL["repeat"]

# ################ Repeat class ###############################################


class Repeat:

    """ Tandem repeats are consecutive copies of genomic sequence.

    A tandem repeat is stored in terms of a multiple sequence alignment of its
    tandem repeat units. For example, the amino acid tandem repeat
    KLMKLMKLKLM
    can be displayed as a multiple sequence alignment (here: "multiple repeat
    unit alignment")
    KLM
    KLM
    KL-
    KLM
    and can be stored as
    ["KLM", "KLM", "KL-", "KLM"].
    The - char stands for a gap: It indicates that either, there used to be a
    char at the same position, but it got deleted. Or, all other chars in the
    same column got inserted at some point.

    For research on tandem repeats, many different statistics need to be
    calculated on the multiple sequence alignment of the tandem repeat.

    [Description of these statistics]

    Attributes:
        msa_original (list of str): The original alignment of repeat units.
        msa (list of str):  The alignment of repeat units without rare chars.
        msaT (list of str): The transposed alignment of repeat units. E.g.
            ["KKKK", "LLLL", "MM-M"].
        sequence_type (str): "AA" or "DNA"
        n (int): The number of repeat units in the original msa. E.g. 4.
        l_msa (int): The length of the repeat units in the original msa.
                     E.g. 3.
        l_effective (int): The number of columns in the original msa not
                           counting insertion columns (i.e. columns with more
                           gaps than chars).
        n_effective (float): The number of chars in the MSA without insertion
                             columns divided by n_effective.
        text (str): The repeat region of the tandem repeats (e.g. all chars in
                    the MSA without gaps concatenated.)
        n_gaps (int): The number of gaps in ``msa``
        repeat_region_length (int): The lengths of *text*
    """

    def __str__(self):
        """
            Create string for Repeat instance.
        """
        first_line = ">"
        if hasattr(self, 'begin'):
            first_line += " begin:{0}".format(self.begin)
        if hasattr(self, 'l_effective'):
            first_line += " l_effective:{0}".format(self.l_effective)
        if hasattr(self, 'n'):
            first_line += " n:{0}".format(self.n)
        if hasattr(self, 'd_pvalue') and len(self.d_pvalue) >= 1:
            if CONFIG['scoreslist'][-1] in self.d_pvalue:
                pvalue = self.d_pvalue[CONFIG['scoreslist'][-1]]
                model = CONFIG['scoreslist'][-1]
            else:
                model, pvalue = next(iter(self.d_pvalue.items()))
            try:
                divergence = self.d_divergence[model]
            except:
                divergence = None
            first_line += " pvalue:{0} divergence:{1}" \
                          " type:{2}".format(pvalue, divergence, model)

        return first_line + "\n" + "\n".join(self.msa)

    def create(file, input_format, **kwargs):
        """ Read tandem repeat from file.

        Read tandem repeat from file (currently, only pickle is supported)

        Args:
            format (str):  Currently only "pickle"
            file (str): Path to output file

        .. todo:: Write checks for ``format`` and ``file``.
        .. todo:: Write tests.
        .. todo:: Can repeat_io.read_fasta be replaced by e.g. BioPython seqIO?
        """

        if input_format == 'pickle':
            with open(file, 'rb') as fh:
                return pickle.load(fh)
        elif input_format == 'fasta':
            l_repeat = repeat_io.read_fasta(file, **kwargs)
            return [Repeat(msa=i_msa, sequence_type=i_sequence_type) \
                    for i_msa, i_sequence_type in l_repeat]
        elif input_format == 'simulate':
            l_repeat = repeat_io.random_sequence(return_type='repeat', **kwargs)
            return [Repeat(msa=i_msa, sequence_type=i_sequence_type) \
                    for i_msa, i_sequence_type in l_repeat]
        else:
            raise Exception('file_format is unknown: {}.'.format(file_format))

    def write(self, file, file_format):
        """ Write tandem repeat to file.

        Write tandem repeat to file using one of three formats.

        Args:
            file (str): Path to input file
            file_format (str):  Either "fasta", "pickle" or "stockholm"

        .. todo:: Write checks for ``format`` and ``file``.

        """

        if file_format == 'fasta':
            repeat_io.save_repeat_fasta(self.msa, file)
        elif file_format == 'pickle':
            with open(file, 'wb') as fh:
                pickle.dump(self, fh)
        elif file_format == 'stockholm':
            repeat_io.save_repeat_stockholm(self.msa, file)
        else:
            raise Exception('file_format is unknown: {}.'.format(file_format))

    def score(self, score_type):
        if not hasattr(self, 'd_score') or score_type not in self.d_score:
            self.calculate_scores(scoreslist=[score_type])
        if score_type in self.d_score:
            return self.d_score[score_type]

    def pvalue(self, score_type):
        if not hasattr(self, 'd_pvalue') or score_type not in self.d_pvalue:
            self.calculate_pvalues(scoreslist=[score_type])
        if score_type in self.d_pvalue:
            return self.d_pvalue[score_type]

    def divergence(self, score_type):
        if not hasattr(
                self,
                'd_divergence') or score_type not in self.d_divergence:
            self.calculate_scores(scoreslist=[score_type])
        if score_type in self.d_divergence:
            return self.d_divergence[score_type]

    def __init__(self, msa, begin=None,
                 sequence_type=CONFIG_GENERAL["sequence_type"],
                 scoreslist=CONFIG['scoreslist'],
                 calc_score=CONFIG.as_bool('calc_score'),
                 calc_pvalue=CONFIG.as_bool('calc_pvalue')):
        """
        .. todo:: if calc_score == False and calc_pvalue == True, there is an
            error. However, calc_pvalue == True should automatically require
            the score to be calculated.

        """

        LOG.debug(msa)

        # The index of begin start out on Zero.
        if begin is not None:
            self.begin = begin

        # Replace underscores and asterisks with scores, transform letters to
        # uppercase, replace Selenocysteine (U) with Cysteine (C),
        # replace Pyrrolysine (O) with Lysine (K).
        self.msa = [repeat_unit.upper().replace("_", "-").replace("*", "-")
                    for repeat_unit in msa]

        self.msa_standard_aa = [standardize(i) for i in self.msa]

        self.sequence_type = sequence_type

        self.n = len(self.msa)
        self.l_msa = max(map(len, self.msa))

        # Assure every line is equally long
        for i, iMSA in enumerate(self.msa):
            self.msa[i] += "-" * (self.l_msa - len(iMSA))
        LOG.debug(
            "Constructing repeat from MSA:\n%s",
            "\n".join(self.msa))

        # Transpose MSA
        self.msaT = ["".join(c) for c in zip(*self.msa)]

        # Get text for MSA.
        self.text = " ".join(self.msa)

        self.n_gaps = self.text.count("-")
        self.repeat_region_length = len(
            self.text) - self.n_gaps - self.text.count(" ")

        self.delete_insertion_columns()
        if self.l_effective != 0:
            self.gap_structure()

            if calc_score:
                self.calculate_scores(scoreslist=scoreslist)

            if calc_pvalue:
                self.calculate_pvalues(scoreslist=scoreslist)

    def calc_n_effective(self):
        """ Calculate the effective number of repeat units n_effective

        Calculate the number of letters in in all non-insertion columns,
        divided by <l_effective>

        Args:
            self (Repeat instance)

        """

        if self.l_effective != 0:
            self.n_effective = (
                len("".join(self.msaD)) - self.textD_standard_aa.count('-')) / self.l_effective
        else:
            self.n_effective = 0

    def calc_index_msa(self):
        ''' formerly named: set_begin'''

        # Calculate the transposed MSA with the index in protein sequence
        # instead of letter to be able to compare two overlapping Repeats
        # within the same sequence region.
        # python 2: range returns list by default
        index = list(
            range(
                self.begin + self.repeat_region_length - 1, self.begin - 1, -1))
        # Replace every letter with the index in the protein sequence
        msaI = [[j if j == '-' else index.pop() for j in i]
                for i in self.msa]  # the 'I' stands for index
        # Transpose msaI using list-comprehensions/zip/join magic
        self.msaIT = [[i for i in c if i != '-'] for c in zip(*msaI)]
        self.msaIT = [i for i in self.msaIT if len(i) > 0]

    def calculate_scores(self, scoreslist=CONFIG['scoreslist']):
        """ Calculate scores on a Repeat instance.

        Calculate scores on a Repeat instance for all scores in `scoreslist`.

        Args:
            self (Repeat): An instance of the repeat class.

        Kwargs:
            scoreslist(list of str): A list of the names of the scores (e.g.
             ["phylo_gap01"]) that are calculated on the Repeat instance `self`
        """

        if not hasattr(self, 'd_score'):
            self.d_score = defaultdict(float)

        if not hasattr(self, 'd_divergence'):
            self.d_divergence = defaultdict(float)

        worst_score = {
            'entropy': 1,
            'parsimony': 1,
            'pSim': 0,
            'phylo': 0,
            'phylo_gap01': 0,
            'phylo_gap001': 0}

        if self.l_effective == 0:  # Enter worst score
            self.d_score = {
                iScore: worst_score[iScore] for iScore in scoreslist}
        # Enter None, as a score cannot be calculated if there is just one
        # repeat unit.
        elif self.n < 2:
            self.d_score = {iScore: None for iScore in scoreslist}
        else:   # Calculate score
            if 'entropy' in scoreslist:
                self.d_score['entropy'] = \
                    repeat_score.mean_similarity(self, repeat_score.entropy)
            if 'parsimony' in scoreslist:
                self.d_score['parsimony'] = np.round(
                    repeat_score.mean_similarity(
                        self,
                        repeat_score.parsimony),
                    decimals=CONFIG.as_int('precision'))
            if 'pSim' in scoreslist:
                self.d_score['pSim'] = np.round(
                    repeat_score.mean_similarity(
                        self,
                        repeat_score.pSim),
                    decimals=CONFIG.as_int('precision'))
            if 'phylo' in scoreslist:
                self.d_divergence['phylo'], self.d_score['phylo'] = \
                    repeat_score.phylo_star_topology_local(self)
            if 'phylo_gap01' in scoreslist:
                self.d_divergence['phylo_gap01'], self.d_score['phylo_gap01'] = \
                    repeat_score.phylo_star_topology_local(self,
                                                         gaps='row_wise',
                                                         indel_rate_per_site=0.01)
            if 'phylo_gap01_ignore_trailing_gaps' in scoreslist:
                self.d_divergence['phylo_gap01_ignore_trailing_gaps'], \
                    self.d_score['phylo_gap01_ignore_trailing_gaps'] = \
                    repeat_score.phylo_star_topology_local(self, gaps='ignore_trailing_gaps', indel_rate_per_site=0.01)
            if 'phylo_gap01_ignore_coherent_deletions' in scoreslist:
                self.d_divergence['phylo_gap01_ignore_coherent_deletions'], self.d_score[
                    'phylo_gap01_ignore_coherent_deletions'] = repeat_score.phylo_star_topology_local(self, gaps='ignore_coherent_deletions', indel_rate_per_site=0.01)
            if 'phylo_gap01_ignore_trailing_gaps_and_coherent_deletions' in scoreslist:
                self.d_divergence['phylo_gap01_ignore_trailing_gaps_and_coherent_deletions'], \
                    self.d_score['phylo_gap01_ignore_trailing_gaps_and_coherent_deletions'] = \
                    repeat_score.phylo_star_topology_local(self, gaps='ignore_trailing_gaps_and_coherent_deletions', indel_rate_per_site=0.01)
            if 'phylo_gap001' in scoreslist:
                self.d_divergence['phylo_gap001'], self.d_score['phylo_gap001'] = repeat_score.phylo_star_topology_local(
                    self, gaps='row_wise', indel_rate_per_site=0.001)
            if 'phylo_gap001_ignore_trailing_gaps' in scoreslist:
                self.d_divergence['phylo_gap001_ignore_trailing_gaps'], self.d_score[
                    'phylo_gap001_ignore_trailing_gaps'] = repeat_score.phylo_star_topology_local(self, gaps='ignore_trailing_gaps', indel_rate_per_site=0.001)
            if 'phylo_gap001_ignore_coherent_deletions' in scoreslist:
                self.d_divergence['phylo_gap001_ignore_coherent_deletions'], self.d_score[
                    'phylo_gap001_ignore_coherent_deletions'] = repeat_score.phylo_star_topology_local(self, gaps='ignore_coherent_deletions', indel_rate_per_site=0.001)
            if 'phylo_gap001_ignore_trailing_gaps_and_coherent_deletions' in scoreslist:
                self.d_divergence['phylo_gap001_ignore_trailing_gaps_and_coherent_deletions'], self.d_score[
                    'phylo_gap001_ignore_trailing_gaps_and_coherent_deletions'] = repeat_score.phylo_star_topology_local(self, gaps='ignore_trailing_gaps_and_coherent_deletions', indel_rate_per_site=0.001)

    def calculate_pvalues(self, scoreslist=CONFIG['scoreslist']):
        """ Calculate pvalues on a Repeat instance.

        Calculate pvalues on a Repeat instance for all scores in `scoreslist`.

        Args:
            self (Repeat): An instance of the repeat class.

        Kwargs:
            scoreslist(list of Str): A list of the names of the scores (e.g.
            ["phylo_gap01"]) for which p-Values are calculated on the
            Repeat instance `self`
        """

        if not hasattr(self, 'd_pvalue'):
            self.d_pvalue = defaultdict(int)

        if self.l_effective == 0:  # Enter worst p-value
            self.d_pvalue = {iScore: 1.0 for iScore in scoreslist}
        # Enter None, as a pvalue cannot be calculated if there is just one
        # repeat unit.
        elif self.n < 2:
            self.d_pvalue = {iScore: None for iScore in scoreslist}
        else:   # Calculate score
            if 'entropy' in scoreslist:
                self.d_pvalue['entropy'] = \
                    repeat_pvalue.pvalue_from_empirical_list(self, 'entropy')
            if 'parsimony' in scoreslist:
                self.d_pvalue['parsimony'] = repeat_pvalue.pvalue_pars(self)
            if 'pSim' in scoreslist:
                self.d_pvalue['pSim'] = repeat_pvalue.pvalue_psim(self)
            if 'phylo' in scoreslist:
                self.d_pvalue['phylo'] = \
                    repeat_pvalue.pvalue_from_empirical_list(self, 'phylo')
            if 'phylo_gap' in scoreslist:
                self.d_pvalue['phylo_gap'] = \
                    repeat_pvalue.pvalue_from_empirical_list(self, 'phylo_gap')
            if 'phylo_gap01' in scoreslist:
                self.d_pvalue['phylo_gap01'] = \
                    repeat_pvalue.pvalue_from_empirical_list(self, 'phylo_gap01')
            if 'phylo_gap01_ignore_trailing_gaps' in scoreslist:
                self.d_pvalue['phylo_gap01_ignore_trailing_gaps'] = \
                    repeat_pvalue.pvalue_from_empirical_list(
                        self,
                        'phylo_gap01',
                        self.score('phylo_gap01_ignore_trailing_gaps'))
            if 'phylo_gap01_ignore_coherent_deletions' in scoreslist:
                self.d_pvalue['phylo_gap01_ignore_coherent_deletions'] = \
                    repeat_pvalue.pvalue_from_empirical_list(
                        self,
                        'phylo_gap01',
                        self.score('phylo_gap01_ignore_coherent_deletions'))
            if 'phylo_gap01_ignore_trailing_gaps_and_coherent_deletions' in scoreslist:
                self.d_pvalue['phylo_gap01_ignore_trailing_gaps_and_coherent_deletions'] = \
                    repeat_pvalue.pvalue_from_empirical_list(
                        self,
                        'phylo_gap01',
                        self.score('phylo_gap01_ignore_trailing_gaps_and_coherent_deletions'))
            if 'phylo_gap001' in scoreslist:
                self.d_pvalue['phylo_gap001'] = \
                    repeat_pvalue.pvalue_from_empirical_list(self,
                                                             'phylo_gap001')
            if 'phylo_gap001_ignore_trailing_gaps' in scoreslist:
                self.d_pvalue['phylo_gap001_ignore_trailing_gaps'] = \
                    repeat_pvalue.pvalue_from_empirical_list(
                    self,
                    'phylo_gap001',
                    self.score('phylo_gap001_ignore_trailing_gaps'))
            if 'phylo_gap001_ignore_coherent_deletions' in scoreslist:
                self.d_pvalue['phylo_gap001_ignore_coherent_deletions'] = \
                    repeat_pvalue.pvalue_from_empirical_list(
                    self,
                    'phylo_gap001',
                    self.score('phylo_gap001_ignore_coherent_deletions'))
            if 'phylo_gap001_ignore_trailing_gaps_and_coherent_deletions' in scoreslist:
                self.d_pvalue['phylo_gap001_ignore_trailing_gaps_and_coherent_deletions'] = \
                    repeat_pvalue.pvalue_from_empirical_list(
                    self,
                    'phylo_gap001',
                    self.score('phylo_gap001_ignore_trailing_gaps_and_coherent_deletions'))

    def delete_insertion_columns(self):
        """ Create the tandem repeat attributes `msa`, `masT`, `l` and `n`
            without insertion columns.

        We define insertion columns as columns where there are more or equal
        gaps compared to chars.
        These columns are removed from both `msa` to create `msaD` and `msaT`
        to create `msaTD`.

        """

        # We define insertion columns as columns where there are more or equal
        # gaps compared to chars.
        self.msaTD = [column for column in self.msaT
                      if (2 * column.count('-') < len(column))
                      ]
        self.insertions = re.compile("-+").findall(''.join([
            'p' if (2 * column.count('-') < len(column)) else '-'
            for column in self.msaT
        ])
        )

        self.msaTD_standard_aa = [standardize(i) for i in self.msaTD]

        self.msaD = ["".join(c) for c in zip(*self.msaTD)]

        self.l_effective = len(self.msaTD)
        textD = " ".join(self.msaD)
        self.textD_standard_aa = standardize(textD)
        # totD is used in repeat_score
        self.totD = len(textD) - textD.count('-') - self.n + 1
        if self.l_effective != 0:
            self.n_effective = (
                len("".join(self.msaD)) - textD.count('-')) / self.l_effective
        else:
            self.n_effective = 0

    def gap_structure(self):
        ''' Calculate the number and length of insertions and deletions for
            this tandem repeat.'''

        # Needed: Number and size of insertions, number and size of deletions
        # row_wise: (standard case) column wise coherence is ignored, the
        # indels are summarised over the columns.
        # ignore_trailing_gaps: In addition to the standard case, trailing gaps
        # on either side of the tandem repeat are replaced with letters, and
        # hence ignored.
        # The separation into insertions and deletions comes before ignoring
        # trailing gaps.
        # ignore_coherent_deletions : In addition to the standard case, same
        # size deletions spanning the same columns are only counted once.
        self.insertions = []
        self.deletions = {}
        self.gaps = {}

        # row_wise
        # 1. Detect insertions
        # First, name columns either '-' if they are insertion columns, or 'p',
        #  if they are not.
        # Concetanate the string of 'p's and '-', and double it, to consider
        # insertions in the last column
        # and the first column as associated (= of the same origin)
        # Lastly, detect the length of all insertions.
        insertions = re.compile("-+").findall(2 * (''.join([
            'p' if (2 * column.count('-') < len(column)) else '-'
            for column in self.msaT
        ])))
        insertions = insertions[int(len(insertions) / 2):-1] if len(
            insertions) % 2 == 1 else insertions[int(len(insertions) / 2):]
        self.insertions = [len(i) for i in insertions]
        if self.l_effective == 0:
            self.insertions = [self.l_msa]

        # 2. Detect deletions
        # CHECK this lines. you used msaTD before, but swapped to msaD
        deletions = [(m.start() % self.l_msa, len(m.group()))
                     for m in re.finditer(re.compile("-+"), "".join(self.msaD))]
        self.deletions['row_wise'] = [i[1] for i in deletions]
        self.gaps['row_wise'] = self.insertions + self.deletions['row_wise']

        # ignore_coherent_deletions
        # Remove duplicates from the list of tuples(index of deletion, deletion
        # length)
        deletions_ignore_coherent = {iD: None for iD in deletions}
        self.deletions['ignore_coherent_deletions'] = [i[1]
                                                       for i in deletions_ignore_coherent.keys()]
        self.gaps['ignore_coherent_deletions'] = self.insertions + \
            self.deletions['ignore_coherent_deletions']

        # ignore_trailing_gaps
        # To be precise, we should ignore_trailing_gaps only, if we have
        # ascertained that there were no letters in possible insertion columns
        # before the first deletion column. For now, this case is ignored.
        msaD = deepcopy(self.msaD)

        def sub_trailing_gap(match_object):
            return 'x' * len(match_object.group())
        msaD[0] = re.sub('^-+', sub_trailing_gap, msaD[0])
        msaD[-1] = re.sub('-+$', sub_trailing_gap, msaD[-1])

        # CHECK this lines. you used msaTD before, but swapped to msaD
        # msaTD = ["".join(c) for c in zip(*msaD)]
        deletions = [(m.start() % self.l_msa, len(m.group())) \
                     for m in re.finditer(re.compile("-+"), "".join(msaD))]
        self.deletions['ignore_trailing_gaps'] = [i[1] for i in deletions]
        self.gaps['ignore_trailing_gaps'] = self.insertions + \
            self.deletions['ignore_trailing_gaps']

        ## ignore_trailing_gaps & ignore_coherent_deletions
        deletions_ignore_coherent = {iD: None for iD in deletions}
        self.deletions['ignore_trailing_gaps_and_coherent_deletions'] = [
            i[1] for i in deletions_ignore_coherent.keys()]
        self.gaps['ignore_trailing_gaps_and_coherent_deletions'] = self.insertions + \
            self.deletions['ignore_trailing_gaps_and_coherent_deletions']

    def gap_structure_HMM(self):
        """ Calculate for each column in the transposed multiple repeat unit
            alignment `msaTD`

            *   how many deletions of what length begin in this column?:
                 `deletion_lengths`[`iColumn`] = list of deletion lengths
            *   how many insertions of what length begin in potential insertion
                columns before this column?:
                `insertion_lengths`[`iColumn`] = list of insertion lengths

        Args:
            self (Repeat): An instance of the repeat class.
        """

        # Insertions: In which column (with reference to self.msaTD) do how
        # many insertions of what length start?
        # In which columns are insertions?
        insertions = [
            i for i,
            column in enumerate(
                self.msaT) if (
                2 *
                column.count('-') >= len(column))]

        # Which insertions are neighboring == form a block, and on which index
        # with reference to self.msaTD do they start?
        insertion_blocks = {}
        n_insertion_columns = 0
        while insertions:
            current = insertions.pop(0)
            insertion_blocks[current] = {'indices_self.msaT': [current]}
            while insertions and insertions[0] == insertion_blocks[
                    current]['indices_self.msaT'][-1] + 1:
                insertion_blocks[current][
                    'indices_self.msaT'].append(insertions.pop(0))
            insertion_blocks[current]['index_self.msaTD'] = insertion_blocks[
                current]['indices_self.msaT'][0] - n_insertion_columns
            n_insertion_columns += len(
                insertion_blocks[current]['indices_self.msaT'])

        # If an insertions ranges over the repeat unit border, take the
        # insertion on both parts together:
        if 0 in insertion_blocks.keys() and len(
                self.msaT) - 1 in insertion_blocks.keys():
            insertion_blocks[current]['indices_self.msaT'].extend(
                insertion_blocks[0]['indices_self.msaT'])
            insertion_blocks[current][
                'index_self.msaTD'] = insertion_blocks[0]['index_self.msaTD']
            del insertion_blocks[0]

        # Restructure insertion_blocks: We are only interested in the index
        # with reference to self.msaTD, and the start and end index with
        # reference to self.msaT:
        insertion_blocks = {
            i['index_self.msaTD']: (
                i['indices_self.msaT'][0],
                i['indices_self.msaT'][
                    -1]) for i in insertion_blocks.values()}

        # Go to each insertion block, and see how many insertion it includes,
        # and of what length.
        l = len(self.msaT)
        insertion_lengths = {}
        for iL in range(self.l_effective):
            insertion_lengths[iL] = []
            if iL in insertion_blocks:
                begin = insertion_blocks[iL][0]
                end = insertion_blocks[iL][1]
                length = end - begin + 1
                # In case the insertion blocks bridges over then end of the
                # repeat unit:
                if begin > end:
                    length += l
                print(length)

                repeat_sequence = get_repeat_sequence(self.msa)
                for i in range(self.n):
                    insertion = repeat_sequence[
                        i *
                        l +
                        begin:min(
                            len(repeat_sequence),
                            i *
                            l +
                            begin +
                            length)]
                    insertion_lengths[iL].extend(
                        [len(iI) for iI in re.findall('\w+', insertion)])

                # In case the insertion blocks bridges over then end of the
                # repeat unit:
                if begin > end:
                    insertion = repeat_sequence[:end + 1]
                    insertion_lengths[iL].extend(
                        [len(iI) for iI in re.findall('\w+', insertion)])

        LOG.debug(
            "The insertion_lengths of the current TR are: {0}".format(insertion_lengths))

        # Deletions are much easier to handle than insertions, as they are
        # already in the right coordinate system (self.msaTD)
        # Find all deletions, and note their start index, and their lengths.
        deletions_all = [(m.start(), len(m.group()))
                         for m in re.finditer('-+', ''.join(self.msaD))]
        # Group the deletions with respect to the column in which they start:
        deletion_lengths = {iL: \
            [iD[1] for iD in deletions_all if iD[0] % self.l_effective == iL] \
            for iL in range(self.l_effective)}
        LOG.debug(
            "The deletion_lengths of the current TR are: {0}".format(deletion_lengths))

        return insertion_lengths, deletion_lengths

    def save_original_msa(self, sequence):
        ''' recalculate the original msa, assuming that begin is defined
        correctly (index starting counting on 1.) This function might be
        obsolete when the original MSA is saved from the beginning.'''
        repeat_sequence = sequence[
            self.begin -
            1:self.begin -
            1 +
            self.repeat_region_length]

        count = 0
        self.msa_original = []
        for unit in self.msa:
            unit_original = ''
            for char in unit:
                if char == '-':
                    unit_original += '-'
                else:
                    unit_original += repeat_sequence[count]
                    count += 1
            self.msa_original.append(unit_original)

# Standardize MSA #############################################################


def standardize(blob):

    for original, replacement in CONFIG_GENERAL[
            'ambiguous_amino_acid'].items():
        blob = blob.replace(original, replacement[0])
    return blob


def get_repeat_sequence(msa):

    return "".join([i.replace("_", "").replace("-", "") for i in msa])


# Localize repeat #############################################################


def calculate_position_in_alignment(begin, length, alignment):
    """ Calculate the index of the begin and the end of a TR within an alignment
        returns

        Calculate the index of the begin and the end of a TR within an alignment
        returns

    Args:
        begin (int): (needs explaining)
        length (int): (needs explaining)
        alignment (?): (needs explaining)

    Returns:
        Dict
    """

    # a alignment.seq (hopefully!) is a string of letters, with many gaps "-",
    # as it this particular case is an alignment sequence.
    seq_indexed = [i for i, j in enumerate(str(alignment)) if j != '-']
    LOG.debug("begin: {0}; length: {1}".format(str(begin), str(length)))
    LOG.debug("alignment: {0}".format(str(alignment)))
    return {
        'begin': seq_indexed[
            begin -
            1],
        'end': seq_indexed[
            begin +
            length -
            2]}



# (C) 2013 Elke Schaper

from os.path import join
import re
import logging
import numpy as np
from collections import defaultdict

from tral.repeat import repeat

logger = logging.getLogger(__name__)

'''
Functions to check whether the exon structure and the repeat unit structure are correlated.

'''


################################### Repeat exon structure ################

def get_exon_structure(my_TR, exon_structure):
    '''
    Exon structure parameters:
    # exon_structure = [{'Exon Chr End (bp)': '43478424', 'Exon Chr Start (bp)': '43477252', 'CDS End': '272', 'CDS Start': '1'}, {'Exon Chr End (bp)': '43476658', 'Exon Chr Start (bp)': '43476498', 'CDS End': '433', 'CDS Start': '273'}, {'Exon Chr End (bp)': '43476158', 'Exon Chr Start (bp)': '43476036', 'CDS End': '556', 'CDS Start': '434'}, {'Exon Chr End (bp)': '43475664', 'Exon Chr Start (bp)': '43475564', 'CDS End': '657', 'CDS Start': '557'}, {'Exon Chr End (bp)': '43475416', 'Exon Chr Start (bp)': '43475194', 'CDS End': '880', 'CDS Start': '658'}, {'Exon Chr End (bp)': '43475046', 'Exon Chr Start (bp)': '43474707', 'CDS End': '951', 'CDS Start': '881'}]
    '''

    # Extract the exon structure in protein coordinates
    exon_start_aa = [int(i['CDS Start'])
                     for i in exon_structure if 'CDS Start' in i and i['CDS Start'] != '']
    exon_end_aa = [int(i['CDS End'])
                   for i in exon_structure if 'CDS End' in i and i['CDS End'] != '']

    if not exon_end_aa == []:
        protein_length = exon_end_aa[-1]
    else:
        protein_length = None

    # Extract the TR structure in protein coordinates
    index = my_TR.begin
    tr_start_aa = []
    for iMSA in my_TR.msa:
        unit_length = len(iMSA.replace("-", ""))
        tr_start_aa.append(index)
        index += unit_length

    return exon_start_aa, protein_length, tr_start_aa, index


def get_exon_measures(my_TR, exon_structure):
    '''
    Calculate diverse exon measures

    Exon structure parameters:
    # exon_structure = [{'Exon Chr End (bp)': '43478424', 'Exon Chr Start (bp)': '43477252', 'CDS End': '272', 'CDS Start': '1'}, {'Exon Chr End (bp)': '43476658', 'Exon Chr Start (bp)': '43476498', 'CDS End': '433', 'CDS Start': '273'}, {'Exon Chr End (bp)': '43476158', 'Exon Chr Start (bp)': '43476036', 'CDS End': '556', 'CDS Start': '434'}, {'Exon Chr End (bp)': '43475664', 'Exon Chr Start (bp)': '43475564', 'CDS End': '657', 'CDS Start': '557'}, {'Exon Chr End (bp)': '43475416', 'Exon Chr Start (bp)': '43475194', 'CDS End': '880', 'CDS Start': '658'}, {'Exon Chr End (bp)': '43475046', 'Exon Chr Start (bp)': '43474707', 'CDS End': '951', 'CDS Start': '881'}]
    '''

    exon_start_aa, protein_length, tr_start_aa, exon_stop_aa = get_exon_structure(
        my_TR, exon_structure)
    if exon_start_aa == [] or tr_start_aa == []:
        return None

    nExon = len(
        [i for i in exon_start_aa if i > tr_start_aa[0] and i < tr_start_aa[-1]]) + 1

    nTR_in_Exon_Max = 0
    for iExon_start, iExon_end in zip(exon_start_aa[:-1], exon_start_aa[1:]):
        lTR_Start_in_Exon = [
            iTR_start for iTR_start in tr_start_aa if iTR_start >= iExon_start and iTR_start < iExon_end]
        if len(lTR_Start_in_Exon) > nTR_in_Exon_Max:
            nTR_in_Exon_Max = len(lTR_Start_in_Exon)

    return nExon, nTR_in_Exon_Max


def print_exon_structure(my_TR, exon_structure):
    '''
    Exon structure parameters:
    # exon_structure = [{'Exon Chr End (bp)': '43478424', 'Exon Chr Start (bp)': '43477252', 'CDS End': '272', 'CDS Start': '1'}, {'Exon Chr End (bp)': '43476658', 'Exon Chr Start (bp)': '43476498', 'CDS End': '433', 'CDS Start': '273'}, {'Exon Chr End (bp)': '43476158', 'Exon Chr Start (bp)': '43476036', 'CDS End': '556', 'CDS Start': '434'}, {'Exon Chr End (bp)': '43475664', 'Exon Chr Start (bp)': '43475564', 'CDS End': '657', 'CDS Start': '557'}, {'Exon Chr End (bp)': '43475416', 'Exon Chr Start (bp)': '43475194', 'CDS End': '880', 'CDS Start': '658'}, {'Exon Chr End (bp)': '43475046', 'Exon Chr Start (bp)': '43474707', 'CDS End': '951', 'CDS Start': '881'}]
    '''

    exon_start_aa, tr_start_aa, protein_length = get_exon_structure(
        my_TR, exon_structure)

    if exon_start_aa == [] or tr_start_aa == []:
        return ""

    protein_structure = ""
    for i in range(1, protein_length + 1):
        if i in exon_start_aa:
            if i in tr_start_aa:
                protein_structure += "B"
            else:
                protein_structure += "E"
        elif i in tr_start_aa:
            protein_structure += "x"
        else:
            protein_structure += "_"

    return protein_structure


def main():

    exon_structure = [{'Exon Chr End (bp)': '43478424',
                       'Exon Chr Start (bp)': '43477252',
                       'CDS End': '272',
                       'CDS Start': '1'},
                      {'Exon Chr End (bp)': '43476658',
                       'Exon Chr Start (bp)': '43476498',
                       'CDS End': '433',
                       'CDS Start': '273'},
                      {'Exon Chr End (bp)': '43476158',
                       'Exon Chr Start (bp)': '43476036',
                       'CDS End': '556',
                       'CDS Start': '434'},
                      {'Exon Chr End (bp)': '43475664',
                       'Exon Chr Start (bp)': '43475564',
                       'CDS End': '657',
                       'CDS Start': '557'},
                      {'Exon Chr End (bp)': '43475416',
                       'Exon Chr Start (bp)': '43475194',
                       'CDS End': '880',
                       'CDS Start': '658'},
                      {'Exon Chr End (bp)': '43475046',
                       'Exon Chr Start (bp)': '43474707',
                       'CDS End': '951',
                       'CDS Start': '881'}]

    #[{'Exon Chr End (bp)': '120068186', 'Exon Chr Start (bp)': '120067591', 'CDS End': '500', 'CDS Start': '1'}, {'Exon Chr End (bp)': '120054800', 'Exon Chr Start (bp)': '120054672', 'CDS End': '629', 'CDS Start': '501'}, {'Exon Chr End (bp)': '120053986', 'Exon Chr Start (bp)': '120053709', 'CDS End': '907', 'CDS Start': '630'}, {'Exon Chr End (bp)': '120050255', 'Exon Chr Start (bp)': '120043356', 'CDS End': '1116', 'CDS Start': '908'}]
    exon_structure = my_gene_tree['leafs']['ENSP00000295628']['exon']
    my_TR = my_TR_data['homologs']['ENSP00000295628']['TR']['HMM']

    exon_structure_aa = [[i['CDS Start'], i['CDS End']]
                         for i in exon_structure]



# (C) 2015 Elke Schaper

"""
    :synopsis: Input/output for hmms.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import logging
import os
import re

LOG = logging.getLogger(__name__)

################################## READ HMMMER3  #########################


def read(hmm_filename, id=None):
    """Read HMM file in HMMER3 format.

    HMMER3 file format is described in detail in
    ftp://selab.janelia.org/pub/software/hmmer3/3.0/Userguide.pdf,
    section 8.

    The general file format is::

        HMMER3/b [3.0b2 | June 2009]
        NAME fn3
        ACC PF00041.12
        DESC Fibronectin type III domain
        LENG 86
        ALPH amino
        (...)

        HMM          A       C       D       E       F       G       H       I    (...)    Y
                    m->m    m->i    m->d    i->m    i->i    d->m    d->d
          COMPO   2.70271 4.89246 3.02314 2.64362 3.59817 2.82566 3.74147 3.08574 (...) 3.22607
                  2.68618 4.42225 2.77519 2.73123 3.46354 2.40513 3.72494 3.29354 (...) 3.61503
                  0.00338 6.08833 6.81068 0.61958 0.77255 0.00000       *
          1       3.16986 5.21447 4.52134 3.29953 4.34285 4.18764 4.30886 3.35801 (...) 3.93889 1 - -
                  2.68629 4.42236 2.77530 2.73088 3.46365 2.40512 3.72505 3.29365 (...) 3.61514
                  0.09796 2.38361 6.81068 0.10064 2.34607 0.48576 0.95510
          2       2.70230 5.97353 2.24744 2.62947 5.31433 2.60356 4.43584 4.79731 (...) 4.25623 3 - -
                  2.68618 4.42225 2.77519 2.73123 3.46354 2.40513 3.72494 3.29354 (...) 3.61503
                  0.00338 6.08833 6.81068 0.61958 0.77255 0.48576 0.95510
          (...)
          86      3.03720 5.94099 3.75455 2.96917 5.26587 2.91682 3.66571 4.11840 (...) 4.99111 121 - E
                  2.68618 4.42225 2.77519 2.73123 3.46354 2.40513 3.72494 3.29354 (...) 3.61503
                  0.00227 6.08723       * 0.61958 0.77255 0.00000       *
        //

    .. important:: The probabilities in this format are stored as negative
        natural log probabilities, e.g. -ln(0.25) = 1.38629. The special case
        of 0 probability is stored as ``*`` which in fact means -∞ (minus
        infinity).

    .. todo:: What does this mean: "Store probabilities as log10arithms. [???]"

    Args:
        hmm_filename (str): Path to the file with model data in the HMMER3
            file format.
        id (str, optional): The identifier for the model to be returned.
            E.g. for Pfam the ``id`` can look like this: 'PF00560'.
            If defined, the function returns only the HMM with an identifier
            that matches the provided id. If a model in the file does not have
            an identifier to check against, it is skipped.

    Returns:
        dict: A dictionary of parameters required to initialise the HMM
        including the id, alphabet, emission probabilities and transition
        probabilities.

        Output format::

            {
                'id': 'PF08261.7',
                'letters': ['A', 'C', 'D', ..., 'Y'],
                'COMPO':
                    {'insertion_emissions': [2.68618, 4.42225, ..., 3.61503],
                      'emissions': [2.28205, 5.14899, ..., 1.92022],
                      'transitions': [0.01467, 4.62483, ..., -inf]},
                '1':
                    {'insertion_emissions': [2.68618, 4.42225, ..., 3.61503],
                     'emissions': [1.00089, 4.54999, ..., 5.23581],
                     'transitions': [0.01467, 4.62483, ..., 0.95510]},
                ...
                '8':
                    {'insertion_emissions': [2.68618, 4.42225, ..., 3.61503],
                     'emissions': [4.12723, 5.39816, ..., 4.58094],
                     'transitions': [0.00990, 4.62006, ..., -inf]}
            }

    """

    pat_start_HMMER3 = re.compile(r"HMMER3")
    pat_accession = re.compile(r"ACC\s+([\w\.]+)")
    pat_HMM = re.compile(r"HMM")
    pat_letters = re.compile(r"\w")
    pat_emissions = re.compile(r"[\w\.]+")
    pat_insertions = re.compile(r"[\d\.]+")
    pat_transition = re.compile(r"[\d\.\*]+")
    pat_end = re.compile(r"//")

    size_alphabet = 20

    # Our possible parser states:
    #
    # 0: searching for HMMER3 Tag
    # 0.1 : searching for identifier Tag
    # 1: searching for HMM Tag
    # 2: Skipping line "m->m    m->i    m->d    i->m    i->i    d->m    d->d"
    # 3: searching for emission probabilities OR end of HMM
    # 4: searching for insertion emission probabilities
    # 5: searching for transition probabilities

    state = 0
    with open(hmm_filename, "rt") as infile:

        for i, line in enumerate(infile):
            LOG.debug("Line %s: %s", i, line[0:-1])
            if 0 == state:
                match = pat_start_HMMER3.match(line)
                if match:
                    LOG.debug(" * (0->0.1) Found model start.")
                    state = 0.1

            elif 0.1 == state:
                hmm = {'id': None}
                match = pat_accession.match(line)
                if match:
                    iID = match.group(1)
                    if id:
                        if id in iID:
                            LOG.debug(" * (0.1->1) Found matching"
                                      " identifier.")
                            hmm['id'] = iID
                            state = 1
                        else:
                            LOG.debug(" * (0.1->0) Found identifier which "
                                      "does not the one provided, skipping "
                                      "model.")
                            state = 0
                    else:
                        hmm['id'] = iID
                        state = 1

                # This is done in order to be able to parse HMMs without an
                # accession number, e.g. when created from a Repeat.
                match2 = pat_HMM.match(line)
                if match2:
                    letters = pat_letters.findall(line[3:])
                    if id:
                        LOG.debug(" * (0.1->0) No identifier found to "
                                  "compare to, skipping model.")
                        state = 0
                    else:
                        LOG.debug(" * (0.1->2) No identifier, found"
                                  " HMM start")
                        LOG.debug("Letters: %s", letters)
                        hmm['letters'] = letters
                        state = 2

            elif 1 == state:
                match = pat_HMM.match(line)
                if match:
                    letters = pat_letters.findall(line[3:])
                    LOG.debug(" * (1->2) Found HMM start")
                    LOG.debug("Letters: %s", letters)
                    hmm['letters'] = letters
                    state = 2

            elif 2 == state:
                LOG.debug(" * (2->3) Skipping line")
                state = 3

            elif 3 == state:
                findall = pat_emissions.findall(line)
                if findall:
                    current_hmm_state = findall[0]
                    if current_hmm_state == 'COMPO':
                        string_emissions = findall[1:]
                    else:
                        string_emissions = findall[1:1 + size_alphabet]
                    LOG.debug(" * (3->4) Found emission probabilities")
                    LOG.debug("Current HMM state: %s", current_hmm_state)
                    emissions = [float(i) if i != '*' else -float('inf')
                                 for i in string_emissions]
                    LOG.debug("Emission probabilities: %s", emissions)
                    hmm[current_hmm_state] = {'emissions': emissions}
                    state = 4
                elif pat_end.match(line):
                    if id:
                        LOG.debug(" * (3->TERMINAL) HMM Found and compiled,"
                                  " return HMM.")
                        LOG.info("Yielding {}, stopping".format(hmm['id']))
                        yield hmm
                    else:
                        LOG.debug(" * (3->0) Finished HMM compilation")
                        LOG.info("Yielding {}".format(hmm['id']))
                        yield hmm
                        state = 0
                else:
                    LOG.debug(" * (3->0) Error: No emission line")
                    state = 0

            elif 4 == state:
                findall = pat_insertions.findall(line)
                if findall:
                    LOG.debug(" * (4->5) Found insertion emission"
                              " probabilities")
                    emissions = [float(i) if i != '*' else -float('inf')
                                 for i in findall]
                    LOG.debug("Insertion emission probabilities: %s",
                              emissions)
                    hmm[current_hmm_state]['insertion_emissions'] = emissions
                    state = 5
                else:
                    LOG.debug(" * (4->0) Error: No insertion emission line")
                    state = 0

            elif 5 == state:
                findall = pat_transition.findall(line)
                if findall:
                    LOG.debug(" * (5->3) Found transition probabilities")
                    transitions = [float(i) if i != '*' else -float('inf')
                                   for i in findall]
                    LOG.debug("Transition probabilities: %s", transitions)
                    hmm[current_hmm_state]['transition'] = transitions
                    state = 3
                else:
                    LOG.debug(" * (5->0) Error: No transition line")
                    state = 0


def split_HMMER3_file(hmm_filename, resultdir):
    """ Split HMMER3 models from a single file ``hmm_filename`` into many files
     in ``resultdir``.

    Helper function: split HMMER3 models from a single file ``hmm_filename``
    into many files in ``resultdir``. The models are named after the HMM
    accession.

    Args:
      hmm_filename (str): Path to HMMER3 file.
      resultdir (str): Path to directory where result files are stored.

    """

    pat_start_HMMER3 = re.compile(r"HMMER3")
    pat_accession = re.compile(r"ACC\s+([\w\.]+)")
    tmp_file = os.path.join(resultdir, "tmp.hmm")
    state = 0
    fh = None

    with open(hmm_filename, "rt") as infile:

        for i, line in enumerate(infile):
            LOG.debug("Line %s: %s", i, line[0:-1])
            if 0 == state:
                match = pat_start_HMMER3.match(line)
                if match:
                    if fh:
                        fh.close()
                        os.rename(
                            tmp_file,
                            os.path.join(
                                resultdir,
                                acc +
                                ".hmm"))
                    LOG.debug(" * (0->1) Found HMM start")
                    state = 1
                    fh = open(tmp_file, "w")
                fh.write(line)
            elif 1 == state:
                fh.write(line)
                match = pat_accession.match(line)
                if match:
                    acc = match.group(1).split(".")[0]
                    LOG.debug(" * (1->0) Found accession")
                    state = 0


def read_HMMER_acc_lengths(hmm_filename):
    """Read HMM file in HMMER3 format. Return the PFAM ID and the lengths of
    each model.

    Read HMM file in HMMER3 format. Return the PFAM ID and the lengths of
    each model.

    Args:
      hmm_filename (str): Path to HMMER3 file.

    Returns:
        (dict of str: int): The number of match states for all HMM models in
        ``hmm_filename``.

      ..  todo:: Decide whether this function is needed.

    """

    pat_accession = re.compile(r"ACC\s+([\w\.]+)")
    pat_length = re.compile(r"LENG\s+([\w\.]+)")
    state = 0
    data = {}

    with open(hmm_filename, "rt") as infile:

        for i, line in enumerate(infile):
            LOG.debug("Line %s: %s", i, line[0:-1])
            if 0 == state:
                match = pat_accession.match(line)
                if match:
                    acc = match.group(1)
                    LOG.debug(" * (0->1) Found name")
                    state = 1
            elif 1 == state:
                match = pat_length.match(line)
                if match:
                    data[acc] = int(match.group(1))
                    LOG.debug(" * (1->0) Found length")
                    state = 0

    return data

# def write_HMMER(hmm, hmm_filename):
#
#     ''' Write <hmm> too <hmm_filename> in HMMER3 format.
#        Compare ftp://selab.janelia.org/pub/software/hmmer3/3.0/Userguide.pdf
#        // 8.File formats '''
#
#
#
#
# def read_HMM_smart(hmm_filename, id = None, type = "ACC"):
#
#     ''' Write <hmm> too <hmm_filename> in smart format. '''
#     # Format not decided yet.
